# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard Library Imports
import logging

# Third Party Imports

# Local Imports
from navigate.model.devices.camera.camera_base import CameraBase

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class HamamatsuOrca(CameraBase):
    """HamamatsuOrca camera class.

    This is by default for an Orca Flash 4.0, Fusion, and Lightning cameras.

    **Configuration**::

        hardware:
          camera:
            -
                type: HamamatsuOrca
                serial_number: 302158
            -
                type: HamamatsuOrca
                serial_number: 302159
        ...

        microscopes:
          microscope_name:
            camera:
              hardware:
                name: camera
                type: HamamatsuOrca
                serial_number: 302352
              lightsheet_rolling_shutter_width: 608
              defect_correct_mode: 2.0 # Off: 1.0, On: 2.0
              delay_percent: 10
              pulse_percent: 1
              x_pixels_step: 4
              y_pixels_step: 4
              x_pixels_min: 4
              y_pixels_min: 4
              exposure_time_range:
                min: 1
                max: 1000
                step: 1
              flip_x: False
              flip_y: False
    """

    def __init__(self, microscope_name, device_connection, configuration):
        """Initialize HamamatsuOrca class.

        Parameters
        ----------
        microscope_name : str
            Name of microscope in configuration
        device_connection : object
            Hardware device to connect to
        configuration : multiprocesing.managers.DictProxy
            Global configuration of the microscope
        """
        super().__init__(microscope_name, device_connection, configuration)

        #: dict: Camera parameters
        self.camera_parameters["x_pixels"] = self.max_image_width
        self.camera_parameters["y_pixels"] = self.max_image_height
        self.camera_parameters["x_pixels_min"] = self.min_image_width
        self.camera_parameters["y_pixels_min"] = self.min_image_height
        self.camera_parameters["x_pixels_step"] = self.step_image_width
        self.camera_parameters["y_pixels_step"] = self.step_image_height
        self.camera_parameters["supported_readout_directions"] = ["Top-to-Bottom", "Bottom-to-Top"]
        #: object: Camera controller
        if self.camera_controller.get_property_value("readout_speed"):
            _, speed_max, _ = self.camera_controller.get_property_range("readout_speed")
            if speed_max is not None:
                self.camera_controller.set_property_value("readout_speed", int(speed_max))
                self.camera_parameters["readout_speed"] = int(speed_max)
            else:
                self.camera_controller.set_property_value("readout_speed", 1)
                self.camera_parameters["readout_speed"] = 1

        self.camera_parameters[
            "pixel_size_in_microns"
        ] = self.camera_controller.get_property_value("pixel_width")

        # Values are pulled from the CameraParameters section of the configuration.yml
        # file. Exposure time converted here from milliseconds to seconds.

        self.camera_controller.set_property_value(
            "defect_correct_mode", self.camera_parameters["defect_correct_mode"]
        )
        self.camera_controller.set_property_value(
            "trigger_active", self.camera_parameters["trigger_active"]
        )
        self.camera_controller.set_property_value(
            "trigger_mode", self.camera_parameters["trigger_mode"]
        )
        self.camera_controller.set_property_value(
            "trigger_polarity", self.camera_parameters["trigger_polarity"]
        )
        self.camera_controller.set_property_value(
            "trigger_source", self.camera_parameters["trigger_source"]
        )

        logger.info("HamamatsuOrca Initialized")

    def __del__(self):
        """Delete HamamatsuOrca class."""
        logger.info("HamamatsuOrca Shutdown")

    @property
    def serial_number(self):
        """Get Camera Serial Number

        Returns
        -------
        serial_number : str
            Serial number for the camera.
        """
        return self.camera_controller._serial_number

    # The pass statements get around camera base calls
    @property
    def max_image_width(self):
        """Get maximum image width.

        Returns
        -------
        max_image_width : int
            Maximum image width.
        """
        return self.camera_controller.max_image_width

    @max_image_width.setter
    def max_image_width(self, value):
        """Set maximum image width."""
        pass

    @property
    def min_image_width(self):
        """Get minimum image width.

        Returns
        -------
        min_image_width : int
            Minimum image width.
        """
        return self.camera_controller.min_image_width

    @min_image_width.setter
    def min_image_width(self, value):
        """Set minimum image width.

        Parameters
        ----------
        value : int
            Minimum image width.
        """
        pass

    @property
    def max_image_height(self):
        """Get maximum image height.

        Returns
        -------
        max_image_height : int
            Maximum image height.
        """
        return self.camera_controller.max_image_height

    @max_image_height.setter
    def max_image_height(self, value):
        """Set maximum image height.

        Parameters
        ----------
        value : int
            Maximum image height.
        """
        pass

    @property
    def min_image_height(self):
        """Get minimum image height.

        Returns
        -------
        min_image_height : int
            Minimum image height.
        """
        return self.camera_controller.min_image_height

    @min_image_height.setter
    def min_image_height(self, value):
        """Set minimum image height.

        Parameters
        ----------
        value : int
            Minimum image height.
        """
        pass

    @property
    def step_image_width(self):
        """Get step image width.

        Returns
        -------
        step_image_width : int
            Step image width.
        """
        return self.camera_controller.step_image_width

    @step_image_width.setter
    def step_image_width(self, value):
        """Set step image width.

        Parameters
        ----------
        value : int
            Step image width.
        """
        pass

    @property
    def step_image_height(self):
        """Get step image height.

        Returns
        -------
        step_image_height : int
            Step image height.
        """
        return self.camera_controller.step_image_height

    @step_image_height.setter
    def step_image_height(self, value):
        """Step image height."""

        pass

    def report_settings(self):
        """Print Camera Settings.

        Prints the current camera settings to the console and the log file."""
        params = [
            "defect_correct_mode",
            "sensor_mode",
            "binning",
            "readout_speed",
            "trigger_active",
            "trigger_mode",
            "trigger_polarity",
            "trigger_source",
            "internal_line_interval",
            "image_height",
            "image_width",
            "exposure_time",
        ]
        for param in params:
            print(param, self.camera_controller.get_property_value(param))
            logger.info(f"{param}, {self.camera_controller.get_property_value(param)}")

    def close_camera(self):
        """Close HamamatsuOrca Camera"""
        self.camera_controller.dev_close()

    def set_sensor_mode(self, mode):
        """Set HamamatsuOrca sensor mode.

        Parameters
        ----------
        mode : str
            'Normal' or 'Light-Sheet'
        """
        modes_dict = {"Normal": 1, "Light-Sheet": 12}
        if mode in modes_dict:
            self.camera_controller.set_property_value("sensor_mode", modes_dict[mode])

            # Update minimum image sizes based on scan mode
            (
                self.camera_controller.min_image_width,
                _,
                self.camera_controller.step_image_width,
            ) = self.camera_controller.get_property_range("subarray_hsize")
            (
                self.camera_controller.min_image_height,
                _,
                self.camera_controller.step_image_height,
            ) = self.camera_controller.get_property_range("subarray_vsize")
        else:
            print("Camera mode not supported")
            logger.info("Camera mode not supported")

    def set_readout_direction(self, mode):
        """Set HamamatsuOrca readout direction.

        Parameters
        ----------
            mode : str
                'Top-to-Bottom', 'Bottom-to-Top', 'bytrigger', or 'diverge'.
        """
        readout_direction_dict = {
            "Top-to-Bottom": 1.0,
            "Bottom-to-Top": 2.0,
            "bytrigger": 3.0,
            "diverge": 5.0,
            "Bidirectional": 6.0,
            "Rev. Bidirectional": 7.0,
        }

        if mode in readout_direction_dict:
            self.camera_controller.set_property_value("readout_direction", readout_direction_dict[mode])
        else:
            print("Camera readout direction not supported")
            logger.info("Camera readout direction not supported")

    def calculate_readout_time(self):
        """Calculate duration of time needed to readout an image.

        Calculates the readout time and maximum frame rate according to the camera
        configuration settings.
        Assumes model C13440 with Camera Link communication from Hamamatsu.
        Currently pulling values directly from the camera.

        Returns
        -------
        readout_time : float
            Duration of time needed to readout an image.
        max_frame_rate : float
            Maximum framerate for a given camera acquisition mode.

        TODO: I think self.camera_controller.get_property_value("readout_time") pulls
              out the actual readout_time
              calculated here (i.e. we don't need to do the calculations).
        """
        h = self.camera_controller.get_property_value("readout_time")
        vn = self.camera_controller.get_property_value("subarray_vsize")
        sensor_mode = self.camera_controller.get_property_value("sensor_mode")
        exposure_time = self.camera_controller.get_property_value("exposure_time")
        trigger_source = self.camera_controller.get_property_value("trigger_source")
        trigger_active = self.camera_controller.get_property_value("trigger_active")

        if sensor_mode == 1:
            #  Area sensor mode operation
            if trigger_source == 1:
                # Internal Trigger Source
                max_frame_rate = 1 / ((vn / 2) * h)
                readout_time = exposure_time - ((vn / 2) * h)

            if trigger_active == 1 or 2:
                #  External Trigger Source
                #  Edge == 1, Level == 2
                max_frame_rate = 1 / ((vn / 2) * h + exposure_time + 10 * h)
                readout_time = exposure_time - ((vn / 2) * h + exposure_time + 10 * h)

            if trigger_active == 3:
                #  External Trigger Source
                #  Synchronous Readout == 3
                max_frame_rate = 1 / ((vn / 2) * h + 5 * h)
                readout_time = exposure_time - ((vn / 2) * h + 5 * h)

            readout_time = h
            max_frame_rate = 1.0 / (exposure_time + readout_time)

        elif sensor_mode == 12:
            #  Progressive sensor mode operation
            max_frame_rate = 1 / (exposure_time + (vn + 10) * h)
            readout_time = exposure_time - 1 / (exposure_time + (vn + 10) * h)

        else:
            print(f"Hamamatsu Camera. Sensor mode {sensor_mode} not supported")
            logger.error(f"Hamamatsu Camera. Sensor mode {sensor_mode} not supported")
            max_frame_rate = 0
            readout_time = 0
        return readout_time, max_frame_rate

    def set_exposure_time(self, exposure_time):
        """Set HamamatsuOrca exposure time.

        Note
        ----
            Units of the Hamamatsu API are in seconds.
            Units for the software are in milliseconds. Conversion is done here.

        Parameters
        ----------
        exposure_time : float
            Exposure time in milliseconds.
        """
        exposure_time = exposure_time / 1000
        return self.camera_controller.set_property_value("exposure_time", exposure_time)

    def set_line_interval(self, line_interval_time):
        """Set HamamatsuOrca line interval.

        Parameters
        ----------
        line_interval_time : float
            Line interval duration.
        """
        return self.camera_controller.set_property_value(
            "internal_line_interval", line_interval_time
        )

    def get_line_interval(self):
        """Get HamamatsuOrca line interval.

        Returns
        -------
        line_interval_time : float
            Line interval duration.
        """
        self.line_interval = self.camera_controller.get_property_value(
            "internal_line_interval"
        )

    def set_binning(self, binning_string):
        """Set HamamatsuOrca binning mode.

        Parameters
        ----------
        binning_string : str
            Desired binning properties (e.g., '1x1', '2x2', '4x4', '8x8', '16x16',
            '1x2', '2x4')

        Returns
        -------
        result: bool
            True if successful, False otherwise.
        """
        binning_dict = {
            "1x1": 1,
            "2x2": 2,
            "4x4": 4,
            # '8x8': 8,
            # '16x16': 16,
            # '1x2': 102,
            # '2x4': 204
        }
        if binning_string not in binning_dict.keys():
            logger.debug(f"can't set binning to {binning_string}")
            print(f"can't set binning to {binning_string}")
            return False
        self.camera_controller.set_property_value(
            "binning", binning_dict[binning_string]
        )
        # idx = binning_string.index("x")
        # x_binning = int(binning_string[:idx])
        # y_binning = int(binning_string[idx + 1 :])
        # self.x_pixels = int(self.x_pixels / x_binning)
        # self.y_pixels = int(self.y_pixels / y_binning)

        # should update experiment in controller side
        # self.configuration['experiment']['CameraParameters']['camera_binning'] =
        #   str(self.x_binning) + 'x' + str(self.y_binning)
        return True

    def set_ROI(self, roi_height=2048, roi_width=2048):
        """Change the size of the active region on the camera.

        Parameters
        ----------
        roi_height : int
            Height of active camera region.
        roi_width : int
            Width of active camera region.

        Returns
        -------
        result: bool
            True if successful, False otherwise.
        """
        # Get the Maximum Number of Pixels from the Configuration File
        camera_height = self.max_image_height
        camera_width = self.max_image_width

        if (
            roi_height > camera_height
            or roi_width > camera_width
            or roi_height < 1
            or roi_width < 1
        ):
            logger.debug(f"can't set roi to {roi_width} and {roi_height}")
            return False

        # Calculate Location of Image Edges
        roi_top = (camera_height - roi_height) / 2
        roi_bottom = roi_top + roi_height
        roi_left = (camera_width - roi_width) / 2
        roi_right = roi_left + roi_width

        # Set ROI
        self.x_pixels, self.y_pixels = self.camera_controller.set_ROI(
            roi_left, roi_top, roi_right, roi_bottom
        )

        logger.info(
            "HamamatsuOrca - subarray_hpos, "
            f"{self.camera_controller.get_property_value('subarray_hpos')}"
        )
        logger.info(
            "HamamatsuOrca - subarray_hsize, "
            f"{self.camera_controller.get_property_value('subarray_hsize')}"
        )
        logger.info(
            "HamamatsuOrca - subarray_vpos, "
            f"{self.camera_controller.get_property_value('subarray_vpos')}"
        )
        logger.info(
            "HamamatsuOrca - subarray_vsize, "
            f"{self.camera_controller.get_property_value('subarray_vsize')}"
        )

        return self.x_pixels == roi_width and self.y_pixels == roi_height

    def initialize_image_series(self, data_buffer=None, number_of_frames=100):
        """Initialize HamamatsuOrca image series.

        Parameters
        ----------
        data_buffer : int
            Size of the data to buffer.  Default is None.
        number_of_frames : int
            Number of frames.  Default is 100.
        """
        self.camera_controller.start_acquisition(data_buffer, number_of_frames)
        self.is_acquiring = True

    def close_image_series(self):
        """Close image series.

        Stops the acquisition and sets is_acquiring flag to False.
        """
        self.camera_controller.stop_acquisition()
        self.is_acquiring = False

    def get_new_frame(self):
        """Get frame from HamamatsuOrca camera.

        Returns
        -------
        frame : numpy.ndarray
            Frame from HamamatsuOrca camera.
        """
        return self.camera_controller.get_frames()

    def get_minimum_waiting_time(self):
        """Get minimum waiting time for HamamatsuOrca.

        This function gets the timing information from the camera device
        cyclic_trigger_period, minimum_trigger_blank, minimum_trigger_interval
        'cyclic_trigger_period' if current device is 0.

        According to the documentation, trigger_blank should be bigger than
        trigger_interval.

        Returns
        -------
        trigger_blank : float
            Minimum waiting time.
        """
        # cyclic_trigger =
        #   self.camera_controller.get_property_value('cyclic_trigger_period')
        trigger_blank = self.camera_controller.get_property_value(
            "minimum_trigger_blank"
        )
        # trigger_interval =
        #   self.camera_controller.get_property_value('minimum_trigger_interval')
        return trigger_blank


class HamamatsuOrcaLightning(HamamatsuOrca):
    """HamamatsuOrcaLightning camera class."""

    def __init__(self, microscope_name, device_connection, configuration):
        """Initialize HamamatsuOrcaLightning class.

        Parameters
        ----------
        microscope_name : str
            Name of microscope in configuration
        device_connection : object
            Hardware device to connect to
        configuration : multiprocesing.managers.DictProxy
            Global configuration of the microscope
        """
        HamamatsuOrca.__init__(self, microscope_name, device_connection, configuration)
        self.camera_parameters["supported_readout_directions"] = ["Top-to-Bottom"]

        logger.info("HamamatsuOrcaLightning Initialized")

    def calculate_light_sheet_exposure_time(
        self, full_chip_exposure_time, shutter_width
    ):
        """Calculate light sheet exposure time.

        Parameters
        ----------
        full_chip_exposure_time : float
            Full chip exposure time.
        shutter_width : int
            Shutter width.

        Returns
        -------
        exposure_time : float
            Exposure time.
        camera_line_interval : float
            Camera line interval.
        """

        camera_line_interval = (full_chip_exposure_time / 1000) / (
            (shutter_width + self.y_pixels - 1) / 4
        )

        self.camera_parameters["line_interval"] = camera_line_interval

        exposure_time = camera_line_interval * (shutter_width / 4) * 1000
        return exposure_time, camera_line_interval


class HamamatsuOrcaFire(HamamatsuOrca):
    def __init__(self, microscope_name, device_connection, configuration):
        """Initialize HamamatsuOrcaFire class.

        Parameters
        ----------
        microscope_name : str
            Name of microscope in configuration
        device_connection : object
            Hardware device to connect to
        configuration : multiprocesing.managers.DictProxy
            Global configuration of the microscope
        """
        HamamatsuOrca.__init__(self, microscope_name, device_connection, configuration)
        self.camera_parameters["supported_readout_directions"] = ["Top-to-Bottom", "Bottom-to-Top", "Bidirectional", "Rev. Bidirectional"]
        # get the max image width and height from the camera
        # it seems the camera will reset to its default value when close the connection with the camera.
        self.camera_parameters["x_pixels"] = self.camera_controller.get_property_value("image_width")
        self.camera_parameters["y_pixels"] = self.camera_controller.get_property_value("image_height")

        logger.info("HamamatsuOrcaFire Initialized")

    @property
    def max_image_width(self):
        """Get maximum image width.

        Returns
        -------
        max_image_width : int
            Maximum image width.
        """
        return self.camera_parameters["x_pixels"]

    @max_image_width.setter
    def max_image_width(self, value):
        """Set maximum image width."""
        pass

    @property
    def max_image_height(self):
        """Get maximum image height.

        Returns
        -------
        max_image_height : int
            Maximum image height.
        """
        return self.camera_parameters["y_pixels"]

    @max_image_height.setter
    def max_image_height(self, value):
        """Set maximum image height.

        Parameters
        ----------
        value : int
            Maximum image height.
        """
        pass


# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import tkinter as tk
from tkinter import ttk
import logging

# Third Party Imports

# Local Imports

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class PopUp(tk.Toplevel):
    """Generic Popup Window class

    Creates the popup window based on the root window being passed,
    title that you want the window to have and the size of the window.
    Some important things to consider:
    - Root has to be the main application window to work
    - Name has to be a string
    - Size also has to be a string in the format '600x400+320+180'
    - 600x400 represents the pixel size
    - +320 means 320 pixels from left edge, +180 means 180 pixels from top edge.
    - If a '-' is used instead of '+' it will be from the opposite edge.
    - Top is a boolean that if true means popup will always be on top of other
    windows
    - Transient is a boolean that if true means the main app will not be usable
    until popup is closed
    - The parent frame for any widgets you add to the popup will be retrieved
    with the get_frame() function

    https://stackoverflow.com/questions/28560209/transient-input-window
    Above link is a resource for using popups.
    Some helpful tips of an easy way to access the data inputted by a user into
    the  popup
    Also discusses transience of a popup (whether you can click out of the popup)

    """

    def __init__(self, root, name, size, top=True, transient=True, *args, **kwargs):
        """Initialize Generic Popup Window class

        Parameters
        ----------
        root : tk.Tk
            The main application window
        name : str
            The title of the popup window
        size : str
            The size of the popup window in the format '600x400+320+180'
        top : bool, optional
            If true, the popup will always be on top of other windows, by default True
        transient : bool, optional
            If true, the main application will not be usable until the popup is closed,
            by default True
        *args
            Variable length argument list
        **kwargs
            Arbitrary keyword arguments

        """
        tk.Toplevel.__init__(self)
        # This starts the popup window config, and makes sure that any child widgets
        # can be resized with the window
        self.title(name)
        self.geometry(size)
        # 300x200 pixels, first +320 means 320 pixels from left edge, +180 means 180
        # pixels from top edge

        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)
        self.resizable(tk.FALSE, tk.FALSE)  # Makes it so user cannot resize
        if top is True:
            self.attributes("-topmost", 1)  # Makes it be on top of mainapp when called

        self.protocol("WM_DELETE_WINDOW", self.dismiss)  # Intercepting close button

        # Checks if you want transience
        if transient is True:
            self.transient(root)  # Prevents clicking outside of window
            self.wait_visibility()  # Can't grab until window appears, so we wait
            self.grab_set()  # Ensures any input goes to this window

        # Putting popup frame into toplevel window
        #: ttk.Frame: The parent frame for any widgets you add to the popup
        self.content_frame = ttk.Frame(self)
        self.content_frame.grid(row=0, column=0, sticky=tk.NSEW)

    def showup(self):
        """Display popup as top-level window.

        This function is used to display the popup window as a top-level window.
        """
        self.deiconify()
        self.attributes("-topmost", 1)

    def dismiss(self):
        """Releases control back to main window from popup

        This function is used to release control back to the main window from the
        popup window.
        """
        self.grab_release()  # Ensures input can be anywhere now
        self.destroy()

    # Function so that popup entries can have a parent frame
    def get_frame(self):
        """Returns the parent frame for any widgets you add to the popup

        This function is used to return the parent frame for any widgets you add to the
        popup window.

        Returns
        -------
        ttk.Frame
            The parent frame for any widgets you add to the popup
        """
        return self.content_frame



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard library imports
from tkinter import filedialog, messagebox, Checkbutton, Label
import traceback
import logging

# Third party imports

# Local application imports
from navigate.controller.sub_controllers.gui_controller import GUIController
from navigate.model.features.restful_features import prepare_service


# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class IlastikPopupController(GUIController):
    """Ilastik popup controller"""

    def __init__(self, view, parent_controller, service_url):
        """Initialize the ilastik popup controller

        Parameters
        ----------
        view : IlastikPopupView
            ilastik popup view
        parent_controller : MainController
            main controller
        service_url : str
            url of the navigate_server for ilastik
        """
        super().__init__(view, parent_controller)
        #: str: url of the navigate_server for ilastik
        self.service_url = service_url
        #: str: ilastik project file name
        self.project_filename = None
        #: bool: show segmentation flag
        self.show_segmentation_flag = False
        #: bool: mark position flag
        self.mark_position_flag = False
        #: dict: label dictionary
        self.label_dict = None
        self.showup()

    def load_project(self):
        """Load ilastik project file."""
        filename = filedialog.askopenfilename(
            defaultextension=".ilp", filetypes=[("Ilastik Project File", "*.ilp")]
        )
        try:
            r = prepare_service(self.service_url, project_file=filename)
            message = (
                "There is something wrong when loading the ilastik project file, "
                "please make sure the file exists and is correct!"
            )
        except Exception as e:
            r = None
            message = "Please make sure the navigate_server for ilastik is running!"
            logger.debug(e)
            logger.debug(traceback.format_exc())

        # destroy current labels
        for child in self.label_frame.winfo_children():
            child.destroy()

        if not r:
            self.project_filename = None
            self.project_filename_var.set(
                "Please select an ilastik pixelclassification project file!"
            )
            messagebox.showerror(title="Ilastik Error", message=message)
        else:
            self.project_filename = filename
            r["status"] = [False] * len(r["names"])
            self.label_dict = r
            self.update_project(filename, r)

    def update_project(self, filename, label_dict):
        """Update project file name and labels

        Parameters
        ----------
        filename : str
            project file name
        label_dict : dict
            label dictionary
        """

        self.project_filename_var.set(filename)
        logger.info(f"{filename} is loaded successfully!")

        # redraw new labels
        for i, label_name in enumerate(label_dict["names"]):
            label_widget = Checkbutton(
                self.label_frame, text=label_name, command=self.toggle_label(i)
            )
            label_widget.grid(
                row=1 + i, column=0, pady=(0, 10), padx=(20, 5), sticky="W"
            )
            if label_dict["status"][i]:
                label_widget.select()
            color_block = Label(
                self.label_frame,
                background=label_dict["label_colors"][i],
                width=3,
                height=1,
            )
            color_block.grid(row=1 + i, column=1, pady=(0, 10), padx=(0, 10))

    def toggle_label(self, label_id):
        """Toggle label status

        Parameters
        ----------
        label_id : int
            label id
        """

        def func():
            """toggle label status"""
            self.label_dict["status"][label_id] = not self.label_dict["status"][
                label_id
            ]

        return func

    def toggle_display(self):
        """Toggle display of ilastik segmentation results"""
        self.show_segmentation_flag = not self.show_segmentation_flag

    def toggle_mark_position(self):
        """Toggle mark position of ilastik segmentation results"""
        self.mark_position_flag = not self.mark_position_flag

    def confirm_setting(self):
        """Confirm settings

        Tell the model which labels will be used activate features containing ilastik
        """
        # update ilastik menu status
        ilastik_menu_state = "normal" if self.project_filename else "disabled"
        self.parent_controller.view.menubar.menu_features.entryconfig(
            "Ilastik Segmentation", state=ilastik_menu_state
        )
        # update segmentation mask color map
        self.parent_controller.camera_view_controller.set_mask_color_table(
            self.label_dict["label_colors"]
        )
        # tell model the target label
        if (
            self.show_segmentation_flag is False
            and self.mark_position_flag is False
            or (self.mark_position_flag and True not in self.label_dict["status"])
        ):
            messagebox.showwarning(
                "Warning", message="You haven't select any usage or target label!"
            )
            self.parent_controller.view.menubar.menu_features.entryconfig(
                "Ilastik Segmentation", state="disabled"
            )
        else:
            self.parent_controller.model.update_ilastik_setting(
                **{
                    "display_segmentation": self.show_segmentation_flag,
                    "mark_position": self.mark_position_flag,
                    "target_labels": [
                        i + 1 for i, x in enumerate(self.label_dict["status"]) if x
                    ],
                }
            )
            # close the window
            self.view.popup.dismiss()

    def showup(self, popup_window=None):
        """Show the ilastik popup window

        This function will let the popup window show in front.

        Parameters
        ----------
        popup_window : IlastikPopupView, optional
            ilastik popup window, by default None
        """
        if popup_window is not None:
            self.view = popup_window
        self.view.popup.protocol("WM_DELETE_WINDOW", self.view.popup.dismiss)
        buttons = self.view.get_buttons()
        buttons["load"].configure(command=self.load_project)
        buttons["confirm"].configure(command=self.confirm_setting)

        self.project_filename_var = self.view.get_variables()["project_name"]
        widgets = self.view.get_widgets()
        self.label_frame = widgets["label_frame"]
        show_seg_widget = widgets["show_segmentation"]
        mark_pos_widget = widgets["mark_position"]

        # segmentation
        if self.show_segmentation_flag:
            show_seg_widget.select()
        if self.mark_position_flag:
            mark_pos_widget.select()

        show_seg_widget.configure(command=self.toggle_display)
        mark_pos_widget.configure(command=self.toggle_mark_position)

        if self.project_filename is not None:
            # destroy current labels
            for child in self.label_frame.winfo_children():
                child.destroy()
            self.update_project(self.project_filename, self.label_dict)



import os, platform
from setuptools import setup, find_packages
from setuptools.extension import Extension
import pip
import shutil

is_windows = 'win' in platform.system().lower()
is_linux = 'lin' in platform.system().lower()

print('Operating system: ' + platform.system())
print('Machine architecture: ' + platform.machine())

if not is_linux and not is_windows:
    print(' Operating systems other than Windows or Linux are not supported.')
    quit()

elif is_linux:
    is_arch_aarch64 = 'aarch64' in platform.machine().lower()
    is_arch_x86_64 = 'x86_64' in platform.machine().lower()
    is_arch_i686 = 'i686' in platform.machine().lower()

    if not is_arch_aarch64 and not is_arch_x86_64 and not is_arch_i686:
        print(' Machine architecture is not supported, it must be aarch64, x86_64 or i686.')
        quit()

elif is_windows:
    is_arch_x86_64 = 'amd64' in platform.machine().lower()
    is_32bit = 'x86' in platform.machine().lower()

    if not is_arch_x86_64 and not is_32bit:
        print(' Machine architecture is not supported, it must be amd64 or x86 32bit.')
        quit()

if is_linux:
    print('************************************************************\n')
    print('Pre-install necessary packages  \n')
    print('   sudo apt-get install python3-pip  \n')
    print('   sudo pip3 install numpy  \n')
    print('************************************************************\n')
    print('Build package:     sudo -E python3 setup.py build  \n')
    print('Install package:   sudo -E python3 setup.py install  \n')
    print('Create Wheel dist: sudo -E python3 setup.py sdist bdist_wheel  \n')
    print('Uninstall package: sudo pip3 uninstall pyvcam  \n')
    print('************************************************************\n')
elif is_windows:
    print('************************************************************\n')
    print('Pre-install necessary packages as admin  \n')
    print('   python -m pip install --upgrade pip setuptools wheel numpy  \n')
    print('************************************************************\n')
    print('Build package:     python setup.py build  \n')
    print('Install package:   python setup.py install  \n')
    print('Create Wheel dist: python setup.py sdist bdist_wheel  \n')
    print('Uninstall package: pip uninstall pyvcam  \n')
    print('************************************************************\n')

pvcam_sdk_path = os.environ['PVCAM_SDK_PATH']
import numpy
include_dirs = [numpy.get_include()]

if is_linux:
    extra_compile_args = ['-std=c++11']
    include_dirs.append('{}/include/'.format(pvcam_sdk_path))

    if is_arch_aarch64:
        lib_dirs = ['{}/library/aarch64'.format(pvcam_sdk_path)]
    elif is_arch_x86_64:
        lib_dirs = ['{}/library/x86_64'.format(pvcam_sdk_path)]
    elif is_arch_i686:
        lib_dirs = ['{}/library/i686'.format(pvcam_sdk_path)]

    libs = ['pvcam']

elif is_windows:
    extra_compile_args = []
    include_dirs.append('{}/inc/'.format(pvcam_sdk_path))

    if is_arch_x86_64:
        lib_dirs = ['{}/Lib/amd64'.format(pvcam_sdk_path)]
        libs = ['pvcam64']
    elif is_32bit:
        lib_dirs = ['{}/Lib/i386'.format(pvcam_sdk_path)]
        libs = ['pvcam32']

ext_modules = [Extension('pyvcam.pvc',
                         ['src/pyvcam/pvcmodule.cpp'],
                         extra_compile_args=extra_compile_args,
                         include_dirs=include_dirs,
                         library_dirs=lib_dirs,
                         libraries=libs)]

setup(name='pyvcam',
      version='2.1.5',
      author='Teledyne Photometrics',
      author_email='Steve.Bellinger@Teledyne.com',
      url='https://github.com/Photometrics/PyVCAM',
      description='Python wrapper for PVCAM functionality.',
      packages=['pyvcam'],
      package_dir={'pyvcam': 'src/pyvcam'},
      py_modules=['pyvcam.constants'],
      ext_modules=ext_modules)

# TODO add checks for if a package is already installed and if so don't install it, if it is installed and up to date give option to update or not
# pip.main(['install', 'wxPython'])
# pip.main(['install', 'pyserial'])
# pip.main(['install', 'opencv-python'])
# pip.main(['install', 'git+https://github.com/pearu/pylibtiff.git'])
# os.system('conda install -y -c conda-forge opencv=3.2.0')

print('\n\n*************** Finished ***************\n')



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


#  Standard Library Imports
import logging

# Third Party Imports

# Local Imports
from navigate.model.waveforms import remote_focus_ramp, smooth_waveform

# # Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class RemoteFocusBase:
    """RemoteFocusBase Class - Parent class for Remote Focusing Device."""

    def __init__(self, microscope_name, device_connection, configuration):
        """Initializes the RemoteFocusBase Class.

        Parameters
        ----------
        microscope_name : str
            Name of the microscope.
        device_connection : str
            Connection string for the remote focus device.
        configuration : dict
            Configuration dictionary.
        """

        #: dict: Configuration dictionary.
        self.configuration = configuration

        #: str: Name of the microscope.
        self.microscope_name = microscope_name

        #: dict: Remote focus device parameters.
        self.device_config = configuration["configuration"]["microscopes"][
            microscope_name
        ]["remote_focus_device"]

        #: int: Sample rate of the DAQ.
        self.sample_rate = configuration["configuration"]["microscopes"][
            microscope_name
        ]["daq"]["sample_rate"]

        #: float: Sweep time of the DAQ.
        self.sweep_time = configuration["configuration"]["microscopes"][
            microscope_name
        ]["daq"]["sweep_time"]

        #: float: Camera delay percent.
        self.camera_delay_percent = configuration["configuration"]["microscopes"][
            microscope_name
        ]["camera"]["delay_percent"]

        # Waveform Parameters
        #: float: Remote focus delay.
        self.remote_focus_delay = self.device_config.get("delay_percent", 7.5)

        #: float: Percent smoothing.
        self.percent_smoothing = self.device_config.get("smoothing", 0)

        #: float: Remote focus ramp falling.
        self.remote_focus_ramp_falling = self.device_config["ramp_falling_percent"]

        #: float: Remote focus max voltage.
        self.remote_focus_max_voltage = self.device_config["hardware"]["max"]

        #: float: Remote focus min voltage.
        self.remote_focus_min_voltage = self.device_config["hardware"]["min"]

        #: dict: Waveform dictionary.
        self.waveform_dict = {}

    def __del__(self):
        """Destructor"""
        pass

    def adjust(self, exposure_times, sweep_times, offset=None):
        """Adjusts the remote focus waveform based on the readout time.

        Parameters
        ----------
        exposure_times : dict
            Dictionary of exposure times for each selected channel
        sweep_times : dict
            Dictionary of sweep times for each selected channel

        Returns
        -------
        waveform : numpy.ndarray
            Waveform for the remote focus device.
        """

        self.waveform_dict = dict.fromkeys(self.waveform_dict, None)
        microscope_state = self.configuration["experiment"]["MicroscopeState"]
        waveform_constants = self.configuration["waveform_constants"]
        imaging_mode = microscope_state["microscope_name"]
        zoom = microscope_state["zoom"]
        self.sample_rate = self.configuration["configuration"]["microscopes"][
            self.microscope_name
        ]["daq"]["sample_rate"]

        duty_cycle_wait_duration = (
            float(
                self.configuration["waveform_constants"]
                .get("other_constants", {})
                .get("remote_focus_settle_duration", 0)
            )
            / 1000
        )

        for channel_key in microscope_state["channels"].keys():
            # channel includes 'is_selected', 'laser', 'filter', 'camera_exposure'...
            channel = microscope_state["channels"][channel_key]

            # Only proceed if it is enabled in the GUI
            if channel["is_selected"] is True:

                # Get the Waveform Parameters - Assumes ETL Delay < Camera Delay.
                # Should Assert.
                laser = channel["laser"]
                exposure_time = exposure_times[channel_key]
                self.sweep_time = sweep_times[channel_key]

                samples = int(self.sample_rate * self.sweep_time)

                # Make sure the smoothing results in a waveform of length sweep time
                ps = float(
                    waveform_constants["remote_focus_constants"][self.microscope_name][
                        zoom
                    ][channel["laser"]].get("percent_smoothing", 0.0)
                )
                if ps > 0:
                    self.sweep_time = (self.sweep_time - duty_cycle_wait_duration) / (
                        1 + ps / 100
                    ) + duty_cycle_wait_duration

                # Remote Focus Parameters
                temp = waveform_constants["remote_focus_constants"][imaging_mode][zoom][
                    laser
                ]["amplitude"]
                if temp == "-" or temp == ".":
                    waveform_constants["remote_focus_constants"][imaging_mode][zoom][
                        laser
                    ]["amplitude"] = "0"

                remote_focus_amplitude = float(
                    waveform_constants["remote_focus_constants"][imaging_mode][zoom][
                        laser
                    ]["amplitude"]
                )

                self.remote_focus_delay = float(
                    waveform_constants["remote_focus_constants"][imaging_mode][zoom][
                        laser
                    ]["percent_delay"]
                )
                self.percent_smoothing = float(
                    waveform_constants["remote_focus_constants"][imaging_mode][zoom][
                        laser
                    ]["percent_smoothing"]
                )

                # Validation for when user puts a '-' in spinbox
                temp = waveform_constants["remote_focus_constants"][imaging_mode][zoom][
                    laser
                ]["offset"]
                if temp == "-" or temp == ".":
                    waveform_constants["remote_focus_constants"][imaging_mode][zoom][
                        laser
                    ]["offset"] = "0"

                remote_focus_offset = float(
                    waveform_constants["remote_focus_constants"][imaging_mode][zoom][
                        laser
                    ]["offset"]
                )
                if offset is not None:
                    remote_focus_offset += offset

                # Calculate the Waveforms
                self.waveform_dict[channel_key] = remote_focus_ramp(
                    sample_rate=self.sample_rate,
                    exposure_time=exposure_time,
                    sweep_time=self.sweep_time,
                    remote_focus_delay=self.remote_focus_delay,
                    camera_delay=self.camera_delay_percent,
                    fall=self.remote_focus_ramp_falling,
                    amplitude=remote_focus_amplitude,
                    offset=remote_focus_offset,
                )

                # Smooth the Waveform if specified
                if self.percent_smoothing > 0:
                    self.waveform_dict[channel_key] = smooth_waveform(
                        waveform=self.waveform_dict[channel_key],
                        percent_smoothing=self.percent_smoothing,
                    )[: samples]

                # Clip any values outside of the hardware limits
                self.waveform_dict[channel_key][
                    self.waveform_dict[channel_key] > self.remote_focus_max_voltage
                ] = self.remote_focus_max_voltage
                self.waveform_dict[channel_key][
                    self.waveform_dict[channel_key] < self.remote_focus_min_voltage
                ] = self.remote_focus_min_voltage

        return self.waveform_dict




# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard Library Imports
import logging
import os

# Third Party Imports
import tifffile

# Local Imports
from navigate.config import get_navigate_path

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class CameraBase:
    """CameraBase - Parent camera class."""

    def __init__(self, microscope_name, device_connection, configuration):
        """Initialize CameraBase class.

        Parameters
        ----------
        microscope_name : str
            Name of microscope in configuration
        device_connection : object
            Hardware device to connect to
        configuration : multiprocesing.managers.DictProxy
            Global configuration of the microscope

        Raises
        ------
        NameError
            If microscope name is not in configuration
        """
        if microscope_name not in configuration["configuration"]["microscopes"].keys():
            raise NameError(f"Microscope {microscope_name} does not exist!")

        #: dict: Global configuration of the microscope
        self.configuration = configuration

        #: object: Hardware device to connect to
        self.camera_controller = device_connection

        #: dict: Camera parameters
        self.camera_parameters = self.configuration["configuration"]["microscopes"][
            microscope_name
        ]["camera"]

        #: bool: Whether the camera is currently acquiring
        self.is_acquiring = False

        # Initialize Pixel Information
        #: int: Maximum image width
        self.max_image_width = 2048
        #: int: Maximum image height
        self.max_image_height = 2048
        #: int: Minimum image width
        self.min_image_width = 4
        #: int: Minimum image height
        self.min_image_height = 4
        #: int: Minimum step size for image width.
        self.step_image_width = 4
        #: int: Minimum step size for image height.
        self.step_image_height = 4
        #: int: Number of pixels in the x direction
        self.x_pixels = self.max_image_width
        #: int: Number of pixels in the y direction
        self.y_pixels = self.max_image_height
        self.camera_parameters["x_pixels"] = self.max_image_width
        self.camera_parameters["y_pixels"] = self.max_image_height
        self.camera_parameters["trigger_source"] = 2.0
        self.camera_parameters["readout_speed"] = 1.0
        self.camera_parameters["pixel_size_in_microns"] = 6.5
        self.camera_parameters["trigger_active"] = 1.0
        self.camera_parameters["trigger_mode"] = 1.0
        self.camera_parameters["trigger_polarity"] = 2.0
        self.camera_parameters["supported_readout_directions"] = [
            "Top-to-Bottom",
            "Bottom-to-Top",
            "Bidirectional",
            "Rev. Bidirectional"
        ]

        # Initialize offset and variance maps, if present
        #: np.ndarray: Offset map
        #: np.ndarray: Variance map
        self._offset, self._variance = None, None
        self.get_offset_variance_maps()

    def get_offset_variance_maps(self):
        """Get offset and variance maps from file.

        Returns
        -------
        offset : np.ndarray
            Offset map.
        variance : np.ndarray
            Variance map.

        Raises
        ------
        FileNotFoundError
            If offset or variance map is not found.
        """
        serial_number = self.camera_parameters["hardware"]["serial_number"]
        try:
            map_path = os.path.join(get_navigate_path(), "camera_maps")
            self._offset = tifffile.imread(
                os.path.join(map_path, f"{serial_number}_off.tiff")
            )
            self._variance = tifffile.imread(
                os.path.join(map_path, f"{serial_number}_var.tiff")
            )
        except FileNotFoundError:
            self._offset, self._variance = None, None
        return self._offset, self._variance

    @property
    def offset(self):
        """Return offset map. If not present, load from file.

        Returns
        -------
        offset : np.ndarray
            Offset map.
        """
        if self._offset is None:
            self.get_offset_variance_maps()
        return self._offset

    @property
    def variance(self):
        """Return variance map. If not present, load from file.

        Returns
        -------
        variance : np.ndarray
            Variance map.
        """

        if self._variance is None:
            self.get_offset_variance_maps()
        return self._variance

    def set_readout_direction(self, mode):
        """Set HamamatsuOrca readout direction.

        Parameters
        ----------
        mode : str
            'Top-to-Bottom', 'Bottom-to-Top', 'bytrigger', or 'diverge'.
        """
        logger.debug(f"set camera readout direction to: {mode}")

    def calculate_light_sheet_exposure_time(
        self, full_chip_exposure_time, shutter_width
    ):
        """Convert normal mode exposure time to light-sheet mode exposure time.
        Calculate the parameters for an acquisition

        Parameters
        ----------
        full_chip_exposure_time : float
            Normal mode exposure time.
        shutter_width : int
            Width of light-sheet rolling shutter.

        Returns
        -------
        exposure_time : float
            Light-sheet mode exposure time (ms).
        camera_line_interval : float
            HamamatsuOrca line interval duration (s).
        """

        camera_line_interval = (full_chip_exposure_time / 1000) / (
            shutter_width + self.y_pixels - 1
        )

        self.camera_parameters["line_interval"] = camera_line_interval

        exposure_time = camera_line_interval * shutter_width * 1000
        return exposure_time, camera_line_interval

    def close_camera(self):
        """Close camera."""
        pass

    def get_line_interval(self):
        """Return stored camera line interval.

        Returns
        -------
        line_interval : float
            line interval duration (s).
        """
        return self.camera_parameters.get("line_interval", None)



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

from pathlib import Path
import os
import inspect
import tkinter as tk

from navigate.config.config import get_navigate_path
from navigate.view.custom_widgets.popup import PopUp
from navigate.tools.file_functions import load_yaml_file, save_yaml_file
from navigate.tools.common_functions import combine_funcs, load_module_from_file
from navigate.tools.decorators import AcquisitionMode
from navigate.model.features import feature_related_functions
from navigate.controller.sub_controllers.gui_controller import GUIController
from navigate.view.popups.plugins_popup import PluginsPopup


class PluginsController:
    """Plugins manaager in the controller side"""

    def __init__(self, view, parent_controller):
        """Initialize plugins manager class"""
        #: object: tkinter frame object.
        self.view = view
        #: object: navigate controller.
        self.parent_controller = parent_controller
        #: str: plugins default path.
        self.plugins_path = os.path.join(
            Path(__file__).resolve().parent.parent.parent, "plugins"
        )
        #: dict: installed plugins with GUI
        self.plugins_dict = {}

    def populate_experiment_setting(self):
        """populate experiment values to plugin GUI"""
        for plugin_name in self.plugins_dict:
            try:
                self.plugins_dict[plugin_name].populate_experiment_setting()
            except:
                pass

    def load_plugins(self):
        """Load plugins"""
        plugins = os.listdir(self.plugins_path)
        installed_plugins = dict(
            [(f, os.path.join(self.plugins_path, f)) for f in plugins]
        )
        # load plugins from plugins_config
        plugins_config_path = os.path.join(
            get_navigate_path(), "config", "plugins_config.yml"
        )
        if not os.path.exists(plugins_config_path):
            save_yaml_file(get_navigate_path(), {}, "plugins_config.yml")
        else:
            plugins_config = load_yaml_file(plugins_config_path)
            for plugin_name, plugin_path in plugins_config.items():
                if plugin_path and os.path.exists(plugin_path):
                    installed_plugins[plugin_name] = plugin_path
        for _, plugin_path in installed_plugins.items():
            if not os.path.isdir(plugin_path):
                continue

            # read "plugin_config.yml"
            plugin_config = load_yaml_file(
                os.path.join(plugin_path, "plugin_config.yml")
            )
            if plugin_config is None:
                continue
            plugin_name = plugin_config.get("name", _)
            plugin_file_name = "_".join(plugin_name.lower().split())
            plugin_class_name = "".join(plugin_name.title().split())
            if "view" in plugin_config:
                # verify if "frame_name" and "file_name" are given and correct
                view_file = os.path.join(
                    plugin_path, "view", f"{plugin_file_name}_frame.py"
                )
                controller_file = os.path.join(
                    plugin_path,
                    "controller",
                    f"{plugin_file_name}_controller.py",
                )
                if (
                    os.path.exists(view_file)
                    and os.path.isfile(view_file)
                    and os.path.exists(controller_file)
                    and os.path.isfile(controller_file)
                ):
                    plugin_frame_module = load_module_from_file(
                        f"{plugin_class_name}Frame", view_file
                    )
                    plugin_frame = getattr(
                        plugin_frame_module, f"{plugin_class_name}Frame"
                    )
                    plugin_controller_module = load_module_from_file(
                        f"{plugin_class_name}Controller", controller_file
                    )
                    plugin_controller = getattr(
                        plugin_controller_module, f"{plugin_class_name}Controller"
                    )

                    if plugin_config["view"] == "Popup":
                        # menu
                        self.parent_controller.view.menubar.menu_plugins.add_command(
                            label=plugin_name,
                            command=self.build_popup_window(
                                plugin_name, plugin_frame, plugin_controller
                            ),
                        )
                    else:
                        self.build_tab_window(
                            plugin_name, plugin_frame, plugin_controller
                        )
            # feature
            features_dir = os.path.join(plugin_path, "model", "features")
            if os.path.exists(features_dir):
                features = os.listdir(features_dir)
                for feature in features:
                    feature_file = os.path.join(features_dir, feature)
                    if os.path.isfile(feature_file):
                        module = load_module_from_file(feature, feature_file)
                        for c in dir(module):
                            if inspect.isclass(getattr(module, c)):
                                setattr(
                                    feature_related_functions, c, getattr(module, c)
                                )

            # acquisition mode
            acquisition_modes = plugin_config.get("acquisition_modes", [])
            for acquisition_mode_config in acquisition_modes:
                acquisition_file = os.path.join(
                    plugin_path, acquisition_mode_config["file_name"]
                )
                if os.path.exists(acquisition_file):
                    module = load_module_from_file(
                        acquisition_mode_config["file_name"][:-3], acquisition_file
                    )
                    acquisition_mode = [
                        m
                        for m in dir(module)
                        if isinstance(getattr(module, m), AcquisitionMode)
                    ]
                    if acquisition_mode:
                        self.parent_controller.add_acquisition_mode(
                            acquisition_mode_config["name"],
                            getattr(module, acquisition_mode[0]),
                        )

    def build_tab_window(self, plugin_name, frame, controller):
        """Build tab for a plugin

        Parameters
        ----------
        plugin_name : str
            plugin name.
        frame: object
            navigate frame object
        controller: object
            navigate controller
        """
        plugin_frame = frame(self.view.settings)
        self.view.settings.add(plugin_frame, text=plugin_name, sticky=tk.NSEW)
        plugin_controller = controller(plugin_frame, self.parent_controller)
        controller_name = (
            "__plugin" + "_".join(plugin_name.lower().split()) + "_controller"
        )
        self.plugins_dict[controller_name] = plugin_controller

    def build_popup_window(self, plugin_name, frame, controller):
        """Build popup window for plugins

        Parameters
        ----------
        plugin_name : str
            plugin name.
        frame: object
            navigate frame object
        controller: object
            navigate controller
        """
        controller_name = (
            "__plugin" + "_".join(plugin_name.lower().split()) + "_controller"
        )

        def func(*args, **kwargs):
            if controller_name in self.plugins_dict:
                self.plugins_dict[controller_name].popup.deiconify()
                return
            popup = PopUp(self.view, plugin_name, "+320+180", transient=False)
            popup.configure(bg="white")
            content_frame = popup.get_frame()
            plugin_frame = frame(content_frame)
            plugin_frame.grid(row=0, column=0, sticky=tk.NSEW, padx=10, pady=10)

            plugin_controller = controller(plugin_frame, self.parent_controller)

            plugin_controller.popup = popup
            popup.deiconify()
            self.plugins_dict[controller_name] = plugin_controller

            popup.protocol(
                "WM_DELETE_WINDOW",
                combine_funcs(
                    popup.dismiss, lambda: self.plugins_dict.pop(controller_name)
                ),
            )

        return func


class UninstallPluginController(GUIController):
    def __init__(self, view, parent_controller):
        super().__init__(view, parent_controller)

        self.plugin_config_path = os.path.join(get_navigate_path(), "config")

        self.popup = PluginsPopup(view)
        self.popup.uninstall_btn.config(command=self.uninstall_plugins)
        self.popup.popup.protocol("WM_DELETE_WINDOW", self.exit_func)
        self.refresh_plugins()

    def showup(self):
        """Show up the popup window"""
        self.refresh_plugins()
        self.popup.popup.showup()

    def exit_func(self):
        """Exit"""
        self.popup.popup.dismiss()
        delattr(self.parent_controller, "uninstall_plugin_controller")

    def refresh_plugins(self):
        """Show installed plugins"""
        self.plugin_config = load_yaml_file(
            os.path.join(self.plugin_config_path, "plugins_config.yml")
        )

        self.popup.build_widgets(self.plugin_config)

    def uninstall_plugins(self, *args):
        """Uninstall plugins"""
        feature_lists_path = get_navigate_path() + "/feature_lists"
        features = os.listdir(feature_lists_path)
        feature_config = {}
        for feature_name in features:
            if feature_name.endswith(".yml") and feature_name != "__sequence.yml":
                feature_content = load_yaml_file(os.path.join(feature_lists_path, feature_name))
                if feature_content and feature_content.get("filename", None):
                    feature_config[feature_name] = feature_content["filename"]
        flag = False
        for var in self.popup.variables:
            if var.get():
                # remove the feature list if a deleted plugin has any.
                for feature_name in feature_config:
                    if feature_config[feature_name].startswith(self.plugin_config[var.get()]):
                        os.remove(f"{feature_lists_path}/{feature_name}")
                self.plugin_config.pop(var.get())
                flag = True
        if flag:
            save_yaml_file(
                self.plugin_config_path, self.plugin_config, "plugins_config.yml"
            )
            tk.messagebox.showwarning(
                title="Navigate",
                message="Plugins are uninstalled! Please restart Navigate!",
            )
            self.showup()



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

#  Standard Imports
import os
import logging
import shutil

# Third Party Imports
import numpy as np
from tifffile import imsave

# Local imports
from navigate.model import data_sources

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class ImageWriter:
    """Class for saving acquired data to disk."""

    def __init__(self, model, data_buffer=None, sub_dir="", image_name=None, saving_flags=None, saving_config={}):
        """Class for saving acquired data to disk.

        Parameters
        ----------
        model : navigate.model.model.Model
            Navigate Model class for controlling hardware/acquisition.
        data_buffer: [SharedNDArray]
            data_buffer will use model's default data_buffer if it's not specified
        sub_dir : str
            Sub-directory of self.model.configuration['experiment']
            ['Saving']['save_directory']
            indicating where to save data
        image_name : str
            Name of the image to be saved. If None, a name will be generated
        """
        #: navigate.model.model.Model: Navigate Model class for controlling
        # hardware/acquisition.
        self.model = model

        #: [SharedNDArray]: Data buffer for saving data.
        self.data_buffer = (
            self.model.data_buffer if data_buffer is None else data_buffer
        )

        #: int : Number of frames in the experiment.
        self.number_of_frames = self.model.number_of_frames

        #: array : Array of saving flags
        self.saving_flags = saving_flags

        #: str : Directory for saving data to disk.
        self.save_directory = ""

        #: str : Sub-directory for saving data to disk.
        self.sub_dir = sub_dir

        #: int : Current time point for saving data.
        self.current_time_point = 0

        #: dict : Dictionary of functions to call for each configuration.
        self.config_table = {
            "signal": {},
            "data": {"main": self.save_image, "cleanup": self.close},
        }

        #: bool: Is 32 vs 64-bit file format.
        self.big_tiff = False

        # create the save directory if it doesn't already exist
        self.save_directory = os.path.join(
            self.model.configuration["experiment"]["Saving"]["save_directory"],
            self.sub_dir,
        )
        try:
            if not os.path.exists(self.save_directory):
                try:
                    os.makedirs(self.save_directory)
                    logger.debug(f"Save Directory Created - {self.save_directory}")
                except OSError:
                    logger.debug(
                        f"Unable to Create Save Directory - {self.save_directory}"
                    )
                    self.model.stop_acquisition = True
                    self.model.event_queue.put(
                        "warning",
                        "Unable to Create Save Directory. Acquisition Terminated",
                    )
                    return
        except FileNotFoundError as e:
            logger.debug(f"Unable to Create Save Directory - {self.save_directory}")
            logger.exception(e)

        # create the MIP directory if it doesn't already exist
        #: np.ndarray : Maximum intensity projection image.
        self.mip = None

        #: str : Directory for saving maximum intensity projection images.
        self.mip_directory = os.path.join(self.save_directory, "MIP")
        try:
            if not os.path.exists(self.mip_directory):
                try:
                    os.makedirs(self.mip_directory)
                    logger.debug(f"MIP Directory Created - {self.mip_directory}")
                except OSError:
                    logger.debug(
                        f"Unable to Create MIP Directory - {self.mip_directory}"
                    )
                    self.model.stop_acquisition = True
                    self.model.event_queue.put(
                        "warning",
                        "Unable to create MIP Directory. Acquisition Terminated.",
                    )
                    return
        except FileNotFoundError as e:
            logger.debug("Image Writer: Unable to create MIP directory.")
            logger.exception(e)

        # Set up the file name and path in the save directory
        #: str : File type for saving data.
        self.file_type = self.model.configuration["experiment"]["Saving"]["file_type"]
        current_channel = self.model.active_microscope.current_channel
        ext = "." + self.file_type.lower().replace(" ", ".").replace("-", ".")
        if image_name is None:
            image_name = self.generate_image_name(current_channel, ext=ext)
        file_name = os.path.join(self.save_directory, image_name)

        # Initialize data source, pointing to the new file name
        #: navigate.model.data_sources.DataSource : Data source for saving data to disk.
        self.data_source = data_sources.get_data_source(self.file_type)(
            file_name=file_name
        )

        # Pass experiment and configuration to metadata
        self.data_source.set_metadata_from_configuration_experiment(
            self.model.configuration
        )

        self.data_source.set_metadata(saving_config)

        # Make sure that there is enough disk space to save the data.
        self.calculate_and_check_disk_space()

        # camera flip flags
        microscope_name = self.model.active_microscope_name
        camera_config = self.model.configuration["configuration"]["microscopes"][
            microscope_name
        ]["camera"]
        self.flip_flags = {
            "x": camera_config.get("flip_x", False),
            "y": camera_config.get("flip_y", False),
        }

    def save_image(self, frame_ids):
        """Save the data to disk.

        Parameters
        ----------
        frame_ids : int
            Index into self.model.data_buffer.
        """

        for idx in frame_ids:

            if (idx < 0) or (idx > (self.number_of_frames - 1)):
                msg = f"Received invalid index {idx}. Skipping this frame."
                logger.debug(msg)
                print(msg)
                continue

            # check the saving flag
            if self.saving_flags:
                if not self.saving_flags[idx]:
                    continue
                self.saving_flags[idx] = False

            # Identify channel, z, time, and position indices
            c_idx, z_idx, t_idx, p_idx = self.data_source._cztp_indices(
                self.data_source._current_frame, self.data_source.metadata.per_stack
            )

            if c_idx == 0 and z_idx == 0:
                # Initialize MIP array with same number of channels as the data
                self.mip = np.ndarray(
                    (
                        int(self.data_source.shape_c),
                        int(self.data_source.shape_y),
                        int(self.data_source.shape_x),
                    )
                ).astype(np.uint16)

            # flip image if necessary
            if self.flip_flags["x"] and self.flip_flags["y"]:
                image = self.data_buffer[idx][::-1, ::-1]
            elif self.flip_flags["x"]:
                image = self.data_buffer[idx][:, ::-1]
            elif self.flip_flags["y"]:
                image = self.data_buffer[idx][::-1, :]
            else:
                image = self.data_buffer[idx]
            # Save data to disk
            try:
                self.data_source.write(
                    image,
                    x=self.model.data_buffer_positions[idx][0],
                    y=self.model.data_buffer_positions[idx][1],
                    z=self.model.data_buffer_positions[idx][2],
                    theta=self.model.data_buffer_positions[idx][3],
                    f=self.model.data_buffer_positions[idx][4],
                )

                # Update MIP
                self.mip[c_idx, :, :] = np.maximum(self.mip[c_idx, :, :], image)

                # Save the MIP
                if (c_idx == self.data_source.shape_c - 1) and (
                    z_idx == self.data_source.shape_z - 1
                ):
                    for c_save_idx in range(self.data_source.shape_c):
                        mip_name = (
                            "P"
                            + str(p_idx).zfill(4)
                            + "_"
                            + "CH0"
                            + str(c_save_idx)
                            + "_"
                            + str(t_idx).zfill(6)
                            + ".tif"
                        )
                        imsave(
                            os.path.join(self.mip_directory, mip_name),
                            self.mip[c_save_idx, :, :],
                        )
            except Exception as e:
                # Close the image, stop the acquisition, log error, and notify user.
                self.close()
                self.model.stop_acquisition = True
                self.model.event_queue.put(("warning", "Insufficient Disk Space. "))
                logger.debug(f"Error - ImageWriter: {e}")
                return

    def generate_image_name(self, current_channel, ext=".tif"):
        """Generates a string for the filename, e.g., CH00_000000.tif.

        Parameters
        ----------
        current_channel : int
            Index into self.model.configuration['experiment']
            ['MicroscopeState']['channels']
            of saved color channel.

        ext : str
            File extension, e.g., '.tif'

        Returns
        -------
        str
            File name, e.g., CH00_000000.tif
        """
        image_name = (
            "CH0"
            + str(current_channel)
            + "_"
            + str(self.current_time_point).zfill(6)
            + ext
        )
        self.current_time_point += 1
        return image_name

    def close(self):
        """Close the data source we are writing to.

        Examples
        --------
        >>> self.close()
        """
        self.data_source.close()

    def calculate_and_check_disk_space(self):
        """Estimate the size of the data that will be written to disk, and confirm
        that sufficient disk space is available. Also evaluates whether
        big-tiff or tiff is needed. Tiff file formats were designed for 32-bit
        operating systems, whereas big-tiff was designed for 64-bit operating systems.

        Assumes 16-bit image type, without compression."""

        # Return disk usage statistics in bytes
        _, _, free = shutil.disk_usage(self.save_directory)

        # Calculate the size in bytes.
        image_size = self.data_source.nbytes

        # Confirm that there is enough disk space to save the data.
        if free < image_size:
            logger.debug("Image Writer: Insufficient Disk Space Estimated.")
            self.model.stop_acquisition = True
            self.model.event_queue.put(
                ("warning", "Insufficient Disk Space. Acquisition Terminated")
            )
            return

        # TIFF vs Big-TIFF Comparison
        if (self.file_type == "TIFF") or (self.file_type == "OME-TIFF"):
            if image_size > 2**32:
                self.data_source.set_bigtiff(True)
            else:
                self.data_source.set_bigtiff(False)



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard library imports

# Third-party imports

# Local application imports


def get_waveform_template_parameters(
    waveform_template_name, waveform_template_dict, microscope_state_dict
):
    """this function will get the parameters of a waveform template

    Parameters
    ----------
    waveform_template_name: str
        the name of the waveform template
    waveform_template_dict: dict
        the dictionary of the waveform templates
    microscope_state_dict: dict
        the dictionary of the microscope state

    Returns
    -------
    repeat_num: int
        the number of repeats
    expand_num: int
        the number of expands
    """
    try:
        waveform_template = waveform_template_dict[waveform_template_name]
    except KeyError:
        repeat_num, expand_num = 1, 1
        return repeat_num, expand_num

    try:
        if type(waveform_template["repeat"]) is int:
            repeat_num = waveform_template["repeat"]
        else:
            repeat_num = int(microscope_state_dict[waveform_template["repeat"]])
    except KeyError:
        repeat_num = 1

    try:
        if type(waveform_template["expand"]) is int:
            expand_num = waveform_template["expand"]
        else:
            expand_num = int(microscope_state_dict[waveform_template["expand"]])
    except KeyError:
        expand_num = 1

    return repeat_num, expand_num



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports

# Third Party Imports
import numpy as np
import matplotlib.ticker as tck
from tkinter import messagebox

# Local Imports
from navigate.controller.sub_controllers.gui_controller import GUIController
from navigate.tools.common_functions import combine_funcs


import logging

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class AutofocusPopupController(GUIController):
    """Class creates the popup to configure autofocus parameters."""

    def __init__(self, view, parent_controller):
        """
        Parameters
        ----------
        view : navigate.view.popups.autofocus_setting_popup.AutofocusPopup
            The view of the autofocus popup.
        parent_controller : navigate.controller.main_controller.MainController
            The parent controller of the autofocus popup.
        """
        super().__init__(view, parent_controller)
        #: dict: The autofocus setting dictionary.
        self.widgets = self.view.get_widgets()
        #: object: The autofocus figure.
        self.autofocus_fig = self.view.fig
        #: object: The autofocus coarse plot.
        self.autofocus_coarse = self.view.coarse
        self.populate_experiment_values()
        #: object: The autofocus coarse plot.
        self.coarse_plot = None

        # add saving function to the function closing the window
        exit_func = combine_funcs(
            self.view.popup.dismiss,
            lambda: delattr(self.parent_controller, "af_popup_controller"),
        )
        self.view.popup.protocol("WM_DELETE_WINDOW", exit_func)
        self.view.autofocus_btn.configure(command=self.start_autofocus)
        self.view.inputs["device"].get_variable().trace_add(
            "write", self.update_device_ref
        )
        self.view.inputs["device_ref"].get_variable().trace_add(
            "write", self.show_autofocus_setting
        )
        for k in self.view.setting_vars:
            self.view.setting_vars[k].trace_add("write", self.update_setting_dict(k))

    def populate_experiment_values(self):
        """Populate Experiment Values

        Populates the experiment values from the experiment settings dictionary
        """
        #: dict: The autofocus setting dictionary.
        self.setting_dict = self.parent_controller.configuration["experiment"][
            "AutoFocusParameters"
        ]
        #: str: The microscope name.
        self.microscope_name = self.parent_controller.configuration["experiment"][
            "MicroscopeState"
        ]["microscope_name"]
        setting_dict = self.setting_dict[self.microscope_name]

        # Default to stages, if they exist.
        if "stage" in setting_dict:
            device = "stage"
        else:
            device = setting_dict.keys()[0]
        self.widgets["device"].widget["values"] = setting_dict.keys()
        self.widgets["device"].set(device)

        # Default to the f axis, if it exists.
        if "f" in setting_dict[device]:
            device_ref = "f"
        else:
            device_ref = setting_dict[device].keys()[0]
        self.widgets["device_ref"].widget["values"] = setting_dict[device].keys()
        self.widgets["device_ref"].set(device_ref)

        for k in self.view.setting_vars:
            self.view.setting_vars[k].set(setting_dict[device][device_ref][k])

    def showup(self):
        """Shows the popup window"""
        self.view.popup.deiconify()
        self.view.popup.attributes("-topmost", 1)

    def start_autofocus(self):
        """Starts the autofocus process."""
        device = self.widgets["device"].widget.get()
        device_ref = self.widgets["device_ref"].widget.get()
        self.parent_controller.configuration["experiment"]["MicroscopeState"][
            "autofocus_device"
        ] = device
        self.parent_controller.configuration["experiment"]["MicroscopeState"][
            "autofocus_device_ref"
        ] = device_ref
        # verify autofocus parameters
        setting_dict = self.setting_dict[self.microscope_name][device][device_ref]
        warning_message = ""
        for k in ["coarse", "fine"]:
            if setting_dict[f"{k}_selected"]:
                try:
                    step = float(setting_dict[f"{k}_step_size"])
                    value = float(setting_dict[f"{k}_range"])
                    if step <= 0 or value < step:
                        warning_message += f"{k} settings are not correct!\n"
                except:
                    warning_message += f"{k} settings are not correct!\n"
        if warning_message:
            messagebox.showerror(
                title="Navigate",
                message=warning_message,
            )
            return
        self.parent_controller.execute("autofocus", device, device_ref)

    def update_device_ref(self, *args):
        """Update device reference name

        Parameters
        ----------
        args: tk event arguments
        """
        device = self.widgets["device"].widget.get()
        device_refs = self.setting_dict[self.microscope_name][device].keys()
        self.widgets["device_ref"].widget["values"] = device_refs
        self.widgets["device_ref"].widget.set(device_refs[0])

    def show_autofocus_setting(self, *args):
        """Show Autofocus Parameters

        Parameters
        ----------
        args: tk event arguments
        """
        device = self.widgets["device"].widget.get()
        device_ref = self.widgets["device_ref"].widget.get()
        setting_dict = self.setting_dict[self.microscope_name]
        for k in self.view.setting_vars:
            self.view.setting_vars[k].set(setting_dict[device][device_ref][k])

    def update_setting_dict(self, parameter):
        """Show Autofocus Parameters

        Parameters
        ----------
        parameter : str
            The parameter to be updated.
        """

        def func(*args):
            device = self.widgets["device"].widget.get()
            device_ref = self.widgets["device_ref"].widget.get()
            self.setting_dict[self.microscope_name][device][device_ref][
                parameter
            ] = self.view.setting_vars[parameter].get()

        return func

    def display_plot(self, data, line_plot=False, clear_data=True):
        """Displays the autofocus plot

        data : numpy.ndarray
            The data to be plotted.
        line_plot : bool
            If True, the plot will be a line plot.
            If False, the plot will be a scatter plot.
        clear_data : bool
            If True, the plot will be cleared before plotting.
            If False, the plot will be added to the existing plot.
        """

        data = np.asarray(data)
        coarse_range = self.setting_dict.get("coarse_range", 500)
        coarse_step = self.setting_dict.get("coarse_step_size", 50)
        fine_range = self.setting_dict.get("fine_range", 50)
        fine_step = self.setting_dict.get("fine_step_size", 5)

        # Calculate the coarse portion of the data
        coarse_steps = int(coarse_range) // int(coarse_step) + 1
        fine_steps = int(fine_range) // int(fine_step) + 1

        if line_plot is True:
            marker = "r-"
        else:
            marker = "k."

        if clear_data is True:
            self.autofocus_coarse.clear()

        # Plotting coarse data
        self.coarse_plot = self.autofocus_coarse.plot(
            data[:coarse_steps, 0], data[:coarse_steps, 1], marker
        )

        # Plotting fine data
        self.coarse_plot = self.autofocus_coarse.plot(
            data[fine_steps:, 0], data[fine_steps:, 1], marker
        )

        # Plot the maxima
        y_max = np.max(data[:, 1])
        peak_loc = data[np.argmax(data[:, 1]), 0]

        y_axes_limit = self.autofocus_coarse.get_ylim()
        x_axes_limit = self.autofocus_coarse.get_xlim()

        # Vertical Indicator
        self.coarse_plot = self.autofocus_coarse.plot(
            [peak_loc, peak_loc], [y_axes_limit[0], y_axes_limit[1]], "--", color="gray"
        )

        # Horizontal Indicator
        self.coarse_plot = self.autofocus_coarse.plot(
            [x_axes_limit[0], x_axes_limit[1]], [y_max, y_max], "--", color="gray"
        )

        # To redraw the plot
        self.autofocus_coarse.set_title("Discrete Cosine Transform", fontsize=18)
        self.autofocus_coarse.set_xlabel("Focus Stage Position", fontsize=16)
        self.autofocus_coarse.ticklabel_format(style="sci", axis="y", scilimits=(0, 0))
        self.autofocus_coarse.yaxis.set_minor_locator(tck.AutoMinorLocator())
        self.autofocus_coarse.xaxis.set_minor_locator(tck.AutoMinorLocator())
        self.autofocus_fig.tight_layout()
        self.autofocus_fig.canvas.draw_idle()



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard Imports
import logging
import time
from threading import Lock

# Third Party Imports

# Local Imports
from navigate.model.devices.daq.daq_base import DAQBase

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class SyntheticDAQ(DAQBase):
    """SyntheticDAQ class for Data Acquisition (DAQ)."""

    def __init__(self, configuration):
        """Initialize the Synthetic DAQ.

        Parameters
        ----------
        configuration : dict
            Configuration dictionary.
        """
        super().__init__(configuration)

        #: dict: Camera object.
        self.camera = {}

        #: Lock: Lock for waiting to run.
        self.wait_to_run_lock = Lock()

        #: dict: Analog output tasks.
        self.analog_outputs = {}

        #: bool: Flag for updating analog task.
        self.is_updating_analog_task = False

        #: str: Trigger mode. Self-trigger or external-trigger.
        self.trigger_mode = "self-trigger"

    def create_camera_task(self):
        """Set up the camera trigger task."""
        pass

    def create_master_trigger_task(self):
        """Set up the DO master trigger task."""
        pass

    def create_galvo_remote_focus_tasks(self):
        """Create galvo and remote focus tasks"""
        pass

    def start_tasks(self):
        """Start the tasks for camera triggering and analog outputs.

        If the tasks are configured to be triggered, they won't start until
        run_tasks() is called."""
        pass

    def stop_tasks(self):
        """Stop the tasks for triggering, analog and counter outputs."""
        pass

    def close_tasks(self):
        """Close the tasks for triggering, analog, and counter outputs."""
        pass

    def prepare_acquisition(self, channel_key):
        """Prepare the acquisition.

        Parameters
        ----------
        channel_key : str
            Channel key for current channel.
        """
        self.current_channel_key = channel_key
        self.is_updating_analog_task = False
        if self.wait_to_run_lock.locked():
            self.wait_to_run_lock.release()

    def run_acquisition(self):
        """Run DAQ Acquisition.

        Run the tasks for triggering, analog and counter outputs.
        The master trigger initiates all other tasks via a shared trigger
        For this to work, all analog output and counter tasks have to be started so that
        they are waiting for the trigger signal."""
        # wait if writing analog tasks
        if self.is_updating_analog_task:
            self.wait_to_run_lock.acquire()
            self.wait_to_run_lock.release()
        time.sleep(0.01)
        if self.trigger_mode == "self-trigger":
            for microscope_name in self.camera:
                self.camera[microscope_name].generate_new_frame()

    def stop_acquisition(self):
        """Stop Acquisition."""
        pass

    def write_waveforms_to_tasks(self):
        """Write the galvo, remote focus, and laser waveforms to each task."""
        pass

    def add_camera(self, microscope_name, camera):
        """Connect camera with daq: only in syntheticDAQ.

        Parameters
        ----------
        microscope_name : str
            Name of microscope.
        camera : Camera
            Camera object.
        """
        self.camera[microscope_name] = camera

    def update_analog_task(self, board_name):
        """Update the analog task.

        Parameters
        ----------
        board_name : str
            Name of board.
        """
        # can't update an analog task while updating one.
        if self.is_updating_analog_task:
            return False

        self.wait_to_run_lock.acquire()
        self.is_updating_analog_task = True

        self.is_updating_analog_task = False
        self.wait_to_run_lock.release()

    def set_external_trigger(self, external_trigger=None):
        """Set the external trigger.

        Parameters
        ----------
        external_trigger : str, optional
            Name of external trigger.
        """

        self.trigger_mode = (
            "self-trigger" if external_trigger is None else "external-trigger"
        )



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import logging

# Third Party Imports
import numpy as np
from scipy import signal

# Local Imports

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


def camera_exposure(sample_rate=100000, sweep_time=0.4, exposure=0.4, camera_delay=10):
    """Calculates timing and duration of camera exposure.
    Not actually used to trigger the camera.  Only meant for visualization.

    Parameters
    ----------
    sample_rate : Integer
        Unit - Hz
    sweep_time : Float
        Unit - Seconds
    exposure : Float
        Unit - Seconds
    camera_delay : Float
        Unit - Percent

    Returns
    -------
    exposure_start : Float
        Unit - Seconds
    exposure_end : Float
        Unit - Seconds

    Examples
    --------
        >>> exposure_start, exposure_end = camera_exposure(sample_rate, sweep_time,
        exposure, camera_delay)

    """
    amplitude = 5

    # get an integer number of samples
    samples = int(np.multiply(sample_rate, sweep_time))

    # create an array just containing the offset voltage:
    array = np.zeros(samples)

    # convert pulse width and delay in % into number of samples
    pulse_delay_samples = int((exposure * camera_delay / 100) * sample_rate)
    pulse_samples = int(exposure * sample_rate)

    # modify the array
    array[pulse_delay_samples : (pulse_samples + pulse_delay_samples)] = amplitude
    return np.array(array)


def single_pulse(
    sample_rate=100000, sweep_time=0.4, delay=10, pulse_width=1, amplitude=1, offset=0
):
    """
    Returns a numpy array with a single pulse
    Used for creating TTL pulses out of analog outputs and laser intensity
    pulses.

    Parameters
    ----------
    sample_rate : Integer
        Unit - Hz
    sweep_time : Float
        Unit - Seconds
    delay : Float
        Unit - Percent
    pulse_width : Float
        Unit - Percent
    amplitude : Float
        Unit - Volts
    offset : Float
        Unit - Volts

    Returns
    -------
    waveform : np.array

    Examples
    --------
    >>> typical_TTL_pulse = single_pulse(sample_rate, sweep_time, 10, 1, 1, 0)
    """
    # get an integer number of samples
    samples = int(np.floor(np.multiply(sample_rate, sweep_time)))

    # create an array just containing the offset voltage:
    array = np.zeros((samples)) + offset

    # convert pulse width and delay in % into number of samples
    pulsedelay_samples = int(samples * delay / 100)
    pulsesamples = int(samples * pulse_width / 100)

    # modify the array
    array[pulsedelay_samples : pulsesamples + pulsedelay_samples] = amplitude
    return np.array(array)


def remote_focus_ramp(
    sample_rate=100000,
    exposure_time=0.2,
    sweep_time=0.24,
    remote_focus_delay=7.5,
    camera_delay=10,
    fall=2.5,
    amplitude=1,
    offset=0,
):
    """Returns a numpy array with a sawtooth ramp - typically used for remote focusing.

    The waveform starts at offset and stays there for the delay period, then
    rises linearly to 2x amplitude (amplitude here refers to 1/2 peak-to-peak)
    and drops back down to the offset voltage during the fall period.

    Switching from a left to right remote focus ramp is possible by exchanging the
    rise and fall periods.

    Parameters
    ----------
    sample_rate : Integer
        Unit - Hz
    exposure_time : Float
        Unit - Seconds
    sweep_time : Float
        Unit - Seconds
    remote_focus_delay : Float
        Unit - Percent
    camera_delay : Float
        Unit - Percent
    fall : Float
        Unit - Percent
    amplitude : Float
        Unit - Volts
    offset : Float
        Unit - Volts

    Returns
    -------
    waveform : np.array

    Examples
    --------
    >>> etl_ramp = tunable_lens_ramp(sample_rate, exposure_time, sweep_time, etl_delay,
        camera_delay, fall, amplitude, offset)

    """

    # create an array just containing the negative amplitude voltage:
    delay_samples = int(remote_focus_delay * exposure_time * sample_rate / 100)
    delay_array = np.zeros(delay_samples) + offset - amplitude

    # 10-7.5 -> 1.025 * .2
    #
    ramp_samples = int(
        (exposure_time + exposure_time * (camera_delay - remote_focus_delay) / 100)
        * sample_rate
    )
    ramp_array = np.linspace(offset - amplitude, offset + amplitude, ramp_samples)

    # fall_samples = .025 * .2 * 100000 = 500
    fall_samples = int(fall * exposure_time * sample_rate / 100)
    fall_array = np.linspace(offset + amplitude, offset - amplitude, fall_samples)

    extra_samples = int(
        int(np.multiply(sample_rate, sweep_time))
        - (delay_samples + ramp_samples + fall_samples)
    )
    if extra_samples > 0:
        extra_array = np.zeros(extra_samples) + offset - amplitude
        waveform = np.hstack([delay_array, ramp_array, fall_array, extra_array])
    else:
        waveform = np.hstack([delay_array, ramp_array, fall_array])

    return waveform


def sawtooth(
    sample_rate=100000,
    sweep_time=0.4,
    frequency=10,
    amplitude=1,
    offset=0,
    duty_cycle=50,
    phase=np.pi / 2,
):
    """
    Returns a numpy array with a sawtooth function.
    Used for creating the galvo signal.

    Parameters
    ----------
    sample_rate : Integer
        Unit - Hz
    sweep_time : Float
        Unit - Seconds
    frequency : Float
        Unit - Hz
    amplitude : Float
        Unit - Volts
    offset : Float
        Unit - Volts
    duty_cycle : Float
        Unit - Percent
    phase : Float
        Unit - Radians

    Returns
    -------
    waveform : np.array

    Examples
    --------
    >>> typical_galvo = sawtooth(sample_rate, sweep_time, 10, 1, 0, 50, np.pi/2)
    """

    samples = int(np.multiply(sample_rate, sweep_time))
    duty_cycle = duty_cycle / 100
    t = np.linspace(0, sweep_time, samples)
    waveform = signal.sawtooth(2 * np.pi * frequency * (t - phase), width=duty_cycle)
    waveform = amplitude * waveform + offset

    return waveform


def dc_value(sample_rate=100000, sweep_time=0.4, amplitude=1):
    """
    Returns a numpy array with a DC value
    Used for creating the resonant galvo drive voltage.

    Parameters
    ----------
    sample_rate : Integer
        Unit - Hz
    sweep_time : Float
        Unit - Seconds
    amplitude : Float
        Unit - Volts

    Returns
    -------
    waveform : np.array

    Examples
    --------
    >>> typical_galvo = dc_value(sample_rate, sweep_time, 1)

    """
    samples = np.multiply(float(sample_rate), sweep_time)
    waveform = np.zeros(int(samples))
    waveform[:] = amplitude
    return waveform


def square(
    sample_rate=100000,
    sweep_time=0.4,
    frequency=10,
    amplitude=1,
    offset=0,
    duty_cycle=50,
    phase=np.pi,
):
    """Returns a numpy array with a square function.
    Used for creating analog laser drive voltage.

    Parameters
    ----------
    sample_rate : Integer
        Unit - Hz
    sweep_time : Float
        Unit - Seconds
    frequency : Float
        Unit - Hz
    amplitude : Float
        Unit - Volts
    offset : Float
        Unit - Volts
    duty_cycle : Float
        Unit - Percent
    phase : Float
        Unit - Radians

    Returns
    -------
    waveform : np.array

    Examples
    --------
    >>> typical_laser = square(sample_rate, sweep_time, 10, 1, 0, 50, np.pi)
    """
    samples = int(sample_rate * sweep_time)
    duty_cycle = duty_cycle / 100
    t = np.linspace(0, sweep_time, samples)
    waveform = signal.square(2 * np.pi * frequency * t + phase, duty=duty_cycle)
    waveform = amplitude * waveform + offset
    return waveform


def sine_wave(
    sample_rate=100000, sweep_time=0.4, frequency=10, amplitude=1, offset=0, phase=0
):
    """Returns a numpy array with a sine waveform

    Used for creating analog laser drive voltage.

    Parameters
    ----------
    sample_rate : int, optional
        Unit - Hz, by default 100000
    sweep_time : float, optional
        Unit - Seconds, by default 0.4
    frequency : int, optional
        Unit - Hz, by default 10
    amplitude : float, optional
        Unit - Volts, by default 1
    offset : float, optional
        Unit - Volts, by default 0
    phase : float, optional
        Unit - Radians, by default 0

    Returns
    -------
    waveform : np.array

    Examples
    --------
    >>> typical_laser = sine_wave(sample_rate, sweep_time, 10, 1, 0, 0)

    """
    samples = int(sample_rate * sweep_time)
    t = np.linspace(0, sweep_time, samples)
    waveform = amplitude * np.sin((2 * np.pi * frequency * t) - phase) + offset
    return waveform


def smooth_waveform(waveform, percent_smoothing=10):
    """Smooths a numpy array via convolution

    Parameters
    ----------
    waveform : np.array
        The waveform to be smoothed
    percent_smoothing : int
        The percentage of the waveform to be smoothed

    Returns
    -------
    smoothed_waveform : np.array
        The smoothed waveform

    Examples
    --------
    >>> smoothed_waveform = smooth_waveform(waveform, percent_smoothing)

    """
    waveform_length = np.size(waveform)
    window_length = int(np.ceil(waveform_length * percent_smoothing / 100))
    if window_length == 0:
        # cannot smooth
        return waveform
    waveform_padded = np.pad(waveform, window_length, mode="edge")
    smoothed_waveform = (
        np.convolve(waveform_padded, np.ones(window_length), "valid") / window_length
    )

    return smoothed_waveform



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import tkinter as tk
import logging

# Third Party Imports

# Local Imports

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class Hover(object):
    """Hover that allows for information to be displayed additionally without
    interrupting the GUI

    Each instance of hover is intended to be an attribute of a tk or ttk widget
    (see validated fields or hovermixins), and by default is set only to show error
    messages

    In this case, the parent widget is the Stage Control GUI y position label.
    To instantiate the description, widget.hover.setdescription("Y position of the
    stage")

    Please note: when dealing with LabelInput widgets, be sure to use
    LabelInput.widget.hover.setdescription(), which will target the specific widget
    and not the LabelInput frame

    Examples of proper usage are within stage_control_tab.py for both LabelInput and
    regular usage
    """

    def __init__(self, widget=None, text=None, type="free"):
        """Initialize the Hover class

        Initializes attributes and binds events.

        Parameters
        ----------
        widget  : bound widget.
            The widget to which the hover instance is bound, usually the one on which
            information is being provided
        text    : str variable
            Text to be displayed when the hover is shown (default set to None so the
            hover will not show at all)
        type    : str variable
            Represents the current state of the hover and whether it is in use at any
            given moment
        """
        #: tk.Widget: The widget to which the hover instance is bound
        self.widget = widget
        #: tk.Toplevel: The hover window
        self.tipwindow = None
        #: int: The id of the widget
        self.id = None
        #: int: The x position of the widget
        #: int: The y position of the widget
        self.x = self.y = 0
        #: str: The text to be displayed when the hover is shown
        self.text = text
        #: str: The current state of the hover
        self.description = None
        #: str: The current state of the hover
        self.type = type

        # define event handling for showing and hiding the hover
        widget.bind("<Enter>", self.show)
        widget.bind("<Leave>", self.hide)
        widget.bind("<ButtonPress>", self.hide)

    # Sets a description for the widget to appear when hovered over
    # If text=None, no description hover will be shown at all
    def setdescription(self, text):
        """Setter for description text

        Parameters
        ----------
        text    : str
            Text to be displayed when hover is shown as a description
        """
        self.description = text

    def getdescription(self):
        """Getter for description text

        Returns
        -------
        type    : str
            The description text
        """
        return self.description

    # Event handlers
    def show(self, event):
        """Event handler to show the hover

        Parameters
        ----------
        event   : event
            The event instance
        """
        if self.type == "free" and self.description is not None:
            self.type = "description"
            self.showtip(self.description)

    def hide(self, event):
        """Event handler to hide the hover

        Parameters
        ----------
        event   : event
            The event instance
        """
        if self.type == "description":
            self.hidetip()

    def update_type(self, newtype):
        """Setter for the type

        Parameters
        ----------
        newtype : str
            The new state of the hover
        """
        self.type = newtype.lower()

    def get_type(self):
        """Getter for the type

        Returns
        -------
        type    : str
            The current state of the hover
        """
        return self.type

    def showtip(self, text):
        """Displays the hover

        Parameters
        ----------
        text    :str
            The text to be displayed on the hover
        """
        self.text = text
        if self.tipwindow or not self.text:
            return

        # set format of hover by type
        if self.type.lower() == "description":
            background = "#ffffe0"
            relief = tk.SOLID
            font = ("tahoma", "8", "normal")
            x = self.widget.winfo_rootx() + self.widget.winfo_width()
            y = self.widget.winfo_rooty() + self.widget.winfo_height()

        elif self.type.lower() == "error":
            background = "#ff5d66"
            relief = (tk.RIDGE,)
            font = ("comic sans", "8", "normal")
            x = self.widget.winfo_rootx()
            y = self.widget.winfo_rooty() + self.widget.winfo_height()

        self.tipwindow = tw = tk.Toplevel(self.widget)
        tw.wm_overrideredirect(1)
        tw.wm_geometry("+%d+%d" % (x, y))
        label = tk.Label(
            tw,
            text=self.text,
            justify=tk.LEFT,
            background=background,
            foreground="black",
            relief=relief,
            borderwidth=1,
            font=font,
        )
        label.pack(ipadx=1)

    def seterror(self, text):
        """Setter for the error message

        Parameters
        ----------
        text    : str
            Error message to be displayed
        """
        self.type = "error"
        self.showtip(text)

    def hidetip(self):
        """Hides the hover and resets type."""
        self.type = "free"
        tw = self.tipwindow
        self.tipwindow = None
        if tw:
            tw.destroy()



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import logging

# Third Party Imports
import nidaqmx
from nidaqmx.errors import DaqError
from nidaqmx.constants import LineGrouping

# Local Imports
from navigate.model.devices.lasers.laser_base import LaserBase

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class LaserNI(LaserBase):
    """LaserNI Class

    This class is used to control a laser connected to a National Instruments DAQ.
    """

    def __init__(self, microscope_name, device_connection, configuration, laser_id):
        """Initialize the LaserNI class.

        Parameters
        ----------
        microscope_name : str
            The microscope name.
        device_connection : object
            The device connection object.
        configuration : dict
            The device configuration.
        laser_id : str
            The laser id.
        """
        super().__init__(microscope_name, device_connection, configuration, laser_id)

        #: str: Modulation type of the laser - Analog or Digital.
        self.on_off_type = None

        # Digital out (if using mixed modulation mode)
        try:
            laser_do_port = self.device_config["onoff"]["hardware"]["channel"]

            #: float: The minimum digital modulation voltage.
            self.laser_min_do = self.device_config["onoff"]["hardware"]["min"]

            #: float: The maximum digital modulation voltage.
            self.laser_max_do = self.device_config["onoff"]["hardware"]["max"]

            #: object: The laser analog or digital modulation task.
            self.laser_do_task = nidaqmx.Task()

            if "/ao" in laser_do_port:
                # Artificial Digital Modulation via an Analog Port
                self.laser_do_task.ao_channels.add_ao_voltage_chan(
                    laser_do_port, min_val=self.laser_min_do, max_val=self.laser_max_do
                )
                self.on_off_type = "analog"

            else:
                # Digital Modulation via a Digital Port
                self.laser_do_task.do_channels.add_do_chan(
                    laser_do_port, line_grouping=LineGrouping.CHAN_FOR_ALL_LINES
                )
                self.on_off_type = "digital"
        except (KeyError, DaqError) as e:
            self.laser_do_task = None
            if isinstance(e, DaqError):
                logger.exception(e)
                logger.debug(e.error_code)
                logger.debug(e.error_type)
                print(e)
                print(e.error_code)
                print(e.error_type)

        #: float: Current laser intensity.
        self._current_intensity = 0

        # Analog out
        try:
            laser_ao_port = self.device_config["power"]["hardware"]["channel"]

            #: float: The minimum analog modulation voltage.
            self.laser_min_ao = self.device_config["power"]["hardware"]["min"]

            #: float: The maximum analog modulation voltage.
            self.laser_max_ao = self.device_config["power"]["hardware"]["max"]

            #: object: The laser analog modulation task.
            self.laser_ao_task = nidaqmx.Task()
            self.laser_ao_task.ao_channels.add_ao_voltage_chan(
                laser_ao_port, min_val=self.laser_min_ao, max_val=self.laser_max_ao
            )
        except DaqError as e:
            logger.exception(e)
            logger.debug(e.error_code)
            logger.debug(e.error_type)
            print(e)
            print(e.error_code)
            print(e.error_type)

    def set_power(self, laser_intensity):
        """Sets the laser power.

        Parameters
        ----------
        laser_intensity : float
            The laser intensity.
        """
        try:
            scaled_laser_voltage = (int(laser_intensity) / 100) * self.laser_max_ao
            self.laser_ao_task.write(scaled_laser_voltage, auto_start=True)
            self._current_intensity = laser_intensity
        except DaqError as e:
            logger.exception(e)

    def turn_on(self):
        """Turns on the laser."""
        try:
            self.set_power(self._current_intensity)
            if self.laser_do_task is not None:
                if self.on_off_type == "digital":
                    self.laser_do_task.write(True, auto_start=True)
                elif self.on_off_type == "analog":
                    self.laser_do_task.write(self.laser_max_do, auto_start=True)
        except DaqError as e:
            logger.exception(e)

    def turn_off(self):
        """Turns off the laser."""
        try:
            tmp = self._current_intensity
            self.set_power(0)
            self._current_intensity = tmp
            if self.laser_do_task is not None:
                if self.on_off_type == "digital":
                    self.laser_do_task.write(False, auto_start=True)
                elif self.on_off_type == "analog":
                    self.laser_do_task.write(self.laser_min_do, auto_start=True)
        except DaqError as e:
            logger.exception(e)

    def close(self):
        """Close the NI Task before exit."""
        try:
            self.laser_ao_task.close()
            if self.laser_do_task is not None:
                self.laser_do_task.close()
        except DaqError as e:
            logger.exception(e)



import cv2
import threading

from pyvcam import pvc
from pyvcam.camera import Camera

NUM_FRAMES = 100

class TriggerThreadRun (threading.Thread):
    def __init__(self, cam):
        threading.Thread.__init__(self)
        self.cam = cam
        self.output = ''

    def run(self):
        cnt = 0

        self.cam.open()
        self.appendOutput('Camera Opened. Name: ' + self.cam.name + ' Sensor Size: ' + str(self.cam.sensor_size))

        width = 400
        height = int(self.cam.sensor_size[1] * width / self.cam.sensor_size[0])
        dim = (width, height)

        self.cam.start_live(exp_time=20)

        while cnt < NUM_FRAMES:

            frame, fps, frame_count = self.cam.poll_frame()

            pixelData = cv2.resize(frame['pixel_data'], dim, interpolation=cv2.INTER_AREA)
            cv2.imshow(self.cam.name, pixelData)
            cv2.waitKey(10)

            if self.output != '':
                self.output += '\n'

            self.appendOutput('CamName:{}\tFrame Rate: {:.1f}\tFrame Count: {:.0f}\tReturned Count: {:.0f}'.format(self.cam.name, fps, frame_count, cnt))
            cnt += 1

        self.cam.finish()
        self.cam.close()

        self.appendOutput('Camera Closed. Name: ' + self.cam.name)

    def appendOutput(self, outputLine):
        if self.output != '':
            self.output += '\n'

        self.output += outputLine

    def getOutput(self):
        ret = self.output
        self.output = ''
        return ret

def main():
    pvc.init_pvcam()

    camera_names = Camera.get_available_camera_names()
    print('Available cameras: ' + str(camera_names))

    threadList = []
    for cameraName in camera_names:
        cam = Camera.select_camera(cameraName)
        thread = TriggerThreadRun(cam)
        thread.start()
        threadList.append(thread)

    checkComplete = False
    while not checkComplete:
        checkComplete = True
        for thread in threadList:
            checkComplete &= not thread.is_alive()

            threadOutput = thread.getOutput()
            if threadOutput != '':
                print(threadOutput)

    print('All cameras complete')
    pvc.uninit_pvcam()

if __name__ == "__main__":
    main()



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import logging
import tkinter as tk
from tkinter import ttk

# Third Party Imports

# Local Imports
from navigate.view.custom_widgets.LabelInputWidgetFactory import LabelInput
from navigate.view.custom_widgets.validation import ValidatedSpinbox, ValidatedEntry

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class CameraSettingsTab(tk.Frame):
    """
    This class holds and controls the layout of the major label frames for the
    camera  settings tab in the settings notebook. Any imported classes are children
    that makeup the content of the major frames. If you need to adjust anything in
    the frames follow the children.
    """

    def __init__(self, setntbk, *args, **kwargs):
        """Initialize the Camera Settings Tab

        Parameters
        ----------
        setntbk : ttk.Notebook
            The settings notebook that the tab will be placed in.
        *args : tuple
            Positional arguments for ttk.Frame
        **kwargs : dict
            Keyword arguments for ttk.Frame
        """
        # Init Frame
        tk.Frame.__init__(self, setntbk, *args, **kwargs)
        #: The index of the tab in the notebook
        self.index = 1

        # Formatting
        tk.Grid.columnconfigure(self, "all", weight=1)
        tk.Grid.rowconfigure(self, "all", weight=1)

        # Camera Modes Frame
        #: tk.Frame: The camera mode frame
        self.camera_mode = CameraMode(self)
        self.camera_mode.grid(row=0, column=0, sticky=(tk.NSEW), padx=10, pady=10)

        # Framerate Label Frame
        #: tk.Frame: The framerate label frame
        self.framerate_info = FramerateInfo(self)
        self.framerate_info.grid(row=0, column=1, sticky=(tk.NSEW), padx=10, pady=10)

        # Region of Interest Settings
        #: tk.Frame: The region of interest settings frame
        self.camera_roi = ROI(self)
        self.camera_roi.grid(
            row=1, column=0, columnspan=2, sticky=(tk.NSEW), padx=10, pady=10
        )


class CameraMode(ttk.Labelframe):
    """This class generates the camera mode label frame.

    Widgets can be adjusted below. Dropdown values need to be set in the controller.
    The function widget.set_values(values) allows this. It can be found in the
    LabelInput class.

    The widgets can be found in the dictionary by using the first word in the label,
    after using get_widgets. The variables tied to each widget can be accessed via
    the widget directly or with the dictionary generated by get_variables.
    """

    def __init__(self, settings_tab, *args, **kwargs):
        """Initialize the Camera Mode Frame

        Parameters
        ----------
        settings_tab : ttk.Frame
            The settings tab that the frame will be placed in.
        *args : tuple
            Positional arguments for ttk.LabelFrame
        **kwargs : dict
            Keyword arguments for ttk.LabelFrame
        """
        # Init Frame
        text_label = "Camera Modes"
        ttk.Labelframe.__init__(self, settings_tab, text=text_label, *args, **kwargs)

        # Formatting
        tk.Grid.columnconfigure(self, "all", weight=1)
        tk.Grid.rowconfigure(self, "all", weight=1)

        # Holds dropdowns, this is done in case more widgets are to be added in a
        # different frame, these can be grouped together
        #: ttk.Frame: The parent frame for any widgets you add.
        content_frame = ttk.Frame(self)
        content_frame.grid(row=0, column=0, sticky=(tk.NSEW), pady=5)

        # Formatting
        tk.Grid.columnconfigure(content_frame, "all", weight=1)
        tk.Grid.rowconfigure(content_frame, "all", weight=1)

        # Dictionary for all the variables, this will be used by the controller
        #: dict: Dictionary of all the widgets in the frame.
        self.inputs = {}
        #: list: List of all the labels for the widgets.
        self.labels = ["Sensor Mode", "Readout Direction", "Number of Pixels"]
        #: list: List of all the names for the widgets.
        self.names = ["Sensor", "Readout", "Pixels"]

        # Dropdown loop
        for i in range(len(self.labels)):
            if i < len(self.labels) - 1:
                self.inputs[self.names[i]] = LabelInput(
                    parent=content_frame,
                    label=self.labels[i],
                    input_class=ttk.Combobox,
                    input_var=tk.StringVar(),
                    input_args={"width": 12},
                )
                self.inputs[self.names[i]].grid(row=i, column=0, pady=3, padx=5)
            else:
                self.inputs[self.names[i]] = LabelInput(
                    parent=content_frame,
                    label=self.labels[i],
                    input_class=ValidatedSpinbox,
                    input_var=tk.StringVar(),
                    input_args={"from_": 0, "to": 10000, "increment": 1, "width": 5},
                )
                self.inputs[self.names[i]].grid(row=i, column=0, pady=3, padx=5)

        # Additional widget settings
        self.inputs["Sensor"].label.grid(padx=(0, 36))
        self.inputs["Readout"].label.grid(padx=(0, 10))
        self.inputs["Pixels"].label.grid(padx=(0, 15))

    def get_variables(self):
        """Get Variables.

        This function returns a dictionary of all the variables that are tied to each
        widget name.

        The key is the widget name, value is the variable associated with the widget.

        Returns
        -------
        variables : dict
            Dictionary of all the variables tied to each widget.
        """
        variables = {}
        for key, widget in self.inputs.items():
            variables[key] = widget.get()
        return variables

    def get_widgets(self):
        """Get Widgets.

        This function returns the dictionary that holds the widgets.

        The key is the widget name, value is the LabelInput class that has all the data.

        Parameters
        ----------
        self.inputs : dict
            Dictionary of all the widgets in the frame.
        """
        return self.inputs


class FramerateInfo(ttk.LabelFrame):
    """Framerate Info Frame.

    This class generates the framerate label frame. Widgets can be adjusted below.
    Entry values need to be set in the controller. The widgets can be found in the
    dictionary by using the first word in the label, after using get_widgets

    The variables tied to each widget can be accessed via the widget directly or with
    the dictionary generated by get_variables.
    """

    def __init__(self, settings_tab, *args, **kwargs):
        """Initialize the Framerate Info Frame

        Parameters
        ----------
        settings_tab : ttk.Frame
            The settings tab that the frame will be placed in.
        *args : tuple
            Positional arguments for ttk.LabelFrame
        **kwargs : dict
            Keyword arguments for ttk.LabelFrame
        """
        # Init Frame
        text_label = "Framerate Info"
        ttk.LabelFrame.__init__(self, settings_tab, text=text_label, *args, **kwargs)

        #  Holds widgets, this is done in case more widgets are to be
        #  added in a different frame, these can be grouped together
        content_frame = ttk.Frame(self)
        content_frame.grid(row=0, column=0, sticky=(tk.NSEW), pady=5, padx=5)

        # Formatting
        tk.Grid.columnconfigure(self, "all", weight=1)
        tk.Grid.rowconfigure(self, "all", weight=1)
        tk.Grid.columnconfigure(content_frame, "all", weight=1)
        tk.Grid.rowconfigure(content_frame, "all", weight=1)

        #  Dictionary for all the variables, this will be used by the controller
        #: dict: Dictionary of all the widgets in the frame.
        self.inputs = {}
        #: list: List of all the labels for the widgets.
        self.labels = [
            "Exposure Time (ms)",
            "Readout Time (ms)",
            "Framerate (Hz)",
            "Images to Average",
        ]
        #: list: List of all the names for the widget values.
        self.names = [
            "exposure_time",
            "readout_time",
            "max_framerate",
            "frames_to_average",
        ]
        #: list: List of all the read only values for the widgets.
        self.read_only = [True, True, True, False]

        #  Dropdown loop
        for i in range(len(self.labels)):
            if self.read_only[i]:
                self.inputs[self.names[i]] = LabelInput(
                    parent=content_frame,
                    label=self.labels[i],
                    input_class=ValidatedEntry,
                    input_var=tk.DoubleVar(),
                    input_args={"width": 6},
                )
                self.inputs[self.names[i]].widget["state"] = "readonly"
                self.inputs[self.names[i]].grid(row=i, column=0, pady=1)
            else:
                self.inputs[self.names[i]] = LabelInput(
                    parent=content_frame,
                    label=self.labels[i],
                    input_class=ValidatedSpinbox,
                    input_var=tk.DoubleVar(),
                    input_args={"from_": 1, "to": 1000, "increment": 1.0, "width": 6},
                )
                self.inputs[self.names[i]].grid(row=i, column=0, pady=1)

        self.inputs["exposure_time"].label.grid(padx=(0, 10))
        self.inputs["readout_time"].label.grid(padx=(0, 14))
        self.inputs["max_framerate"].label.grid(padx=(0, 10))
        self.inputs["frames_to_average"].label.grid(padx=(0, 16))

    def get_variables(self):
        """Get Variables

        This function returns a dictionary of all variables tied to each widget.

        The key is the widget name, value is the variable associated.

        Returns
        -------
        variables : dict
            Dictionary of all the variables tied to each widget.

        """
        variables = {}
        for key, widget in self.inputs.items():
            variables[key] = widget.get()
        return variables

    def get_widgets(self):
        """Get Widgets

        This function returns the dictionary that holds the widgets.
        The key is the widget name, value is the LabelInput class that has all the data.

        Returns
        -------
        self.inputs : dict
            Dictionary of all the widgets in the frame.

        """
        return self.inputs


class ROI(ttk.Labelframe):
    """ROI Frame.

    Widgets can be adjusted below. Dropdown values need to be set in the controller.
    The function widget.set_values(values) allows this. It can be found in the
    LabelInput class. The widgets can be found in the dictionary by using the first
    word in the label, after using get_widgets. The variables tied to each widget can
    be accessed via the widget directly or with the dictionary generated by
    get_variables.
    """

    def __init__(self, settings_tab, *args, **kwargs):
        """Initialize the ROI Frame

        Parameters
        ----------
        settings_tab : ttk.Frame
            The settings tab that the frame will be placed in.
        *args : tuple
            Positional arguments for ttk.LabelFrame
        **kwargs : dict
            Keyword arguments for ttk.LabelFrame
        """
        # Init Frame
        text_label = "Region of Interest Settings"
        ttk.Labelframe.__init__(self, settings_tab, text=text_label, *args, **kwargs)

        # Formatting
        tk.Grid.columnconfigure(self, "all", weight=1)
        tk.Grid.rowconfigure(self, "all", weight=1)

        # Parent Label Frames for widgets
        #: ttk.LabelFrame: The parent frame for any the camera size.
        self.roi_frame = ttk.LabelFrame(self, text="Number of Pixels")
        self.roi_frame.grid(row=0, column=0, sticky=tk.NSEW, padx=10, pady=10)

        # Button Frame
        #: ttk.LabelFrame: The parent frame for default FOV options.
        self.btn_frame = ttk.LabelFrame(self, text="Default FOVs")
        self.btn_frame.grid(row=0, column=1, sticky=tk.NSEW, padx=(40, 10), pady=10)

        # FOV
        #: ttk.LabelFrame: The parent frame for the FOV size.
        self.fov_frame = ttk.LabelFrame(self, text="FOV Dimensions (microns)")
        self.fov_frame.grid(row=1, column=0, sticky=tk.NSEW, padx=10, pady=10)

        # ROI Center
        #: ttk.LabelFrame: The parent frame for the center of the FOV.
        self.center_roi_frame = ttk.LabelFrame(self, text="ROI Center")
        self.center_roi_frame.grid(
            row=1, column=1, sticky=tk.NSEW, padx=(40, 10), pady=10
        )

        # Formatting
        tk.Grid.columnconfigure(self.roi_frame, "all", weight=1)
        tk.Grid.rowconfigure(self.roi_frame, "all", weight=1)
        tk.Grid.columnconfigure(self.btn_frame, "all", weight=1)
        tk.Grid.rowconfigure(self.btn_frame, "all", weight=1)
        tk.Grid.columnconfigure(self.fov_frame, "all", weight=1)
        tk.Grid.rowconfigure(self.fov_frame, "all", weight=1)
        tk.Grid.columnconfigure(self.center_roi_frame, "all", weight=1)
        tk.Grid.rowconfigure(self.center_roi_frame, "all", weight=1)

        # Dictionary for all the variables, this will be used by the controller
        #: dict: Dictionary of all the widgets in the frame.
        self.inputs = {}
        #: dict: Dictionary of all the buttons in the frame.
        self.buttons = {}

        # Labels and names
        #: list: List of all the labels for the widgets.
        self.roi_labels = ["Width", "Height"]  # names are the same, can be reused
        #: list: List of all the names for the widgets.
        self.xy_labels = ["X", "Y"]
        #: list: List of all the names for the widgets.
        self.fov_names = ["FOV_X", "FOV_Y"]
        #: list: List of all the names for the widgets.
        self.center_names = ["Center_X", "Center_Y"]
        #: str: The binning label.
        self.binning = "Binning"

        # Buttons
        btn_labels = ["Use All Pixels", "1600x1600", "1024x1024", "512x512"]
        btn_names = ["All", "1600", "1024", "512"]

        # Loop for each frame
        # Button Frame
        for i in range(len(btn_names)):
            self.buttons[btn_names[i]] = ttk.Button(
                self.btn_frame, text=btn_labels[i], width=9
            )
            self.buttons[btn_names[i]].grid(row=i, column=0, pady=5, padx=35)

        for i in range(2):
            # Num Pix frame
            self.inputs[self.roi_labels[i]] = LabelInput(
                parent=self.roi_frame,
                label=self.roi_labels[i],
                input_class=ValidatedSpinbox,
                input_var=tk.IntVar(),
                input_args={"from_": 0, "increment": 2.0, "width": 5},
            )
            self.inputs[self.roi_labels[i]].grid(row=i, column=0, pady=5, padx=5)

            # FOV Frame
            self.inputs[self.fov_names[i]] = LabelInput(
                parent=self.fov_frame,
                label=self.xy_labels[i],
                input_class=ValidatedSpinbox,
                input_var=tk.IntVar(),
                input_args={"width": 7, "required": True},
            )
            self.inputs[self.fov_names[i]].grid(row=i, column=0, pady=1, padx=5)

            # ROI Center
            self.inputs[self.center_names[i]] = LabelInput(
                parent=self.center_roi_frame,
                label=self.xy_labels[i],
                # input_class=ttk.Spinbox,
                input_class=ValidatedSpinbox,
                input_var=tk.IntVar(),
                input_args={"from_": 0, "to": 2048, "increment": 1.0, "width": 6},
            )
            self.inputs[self.center_names[i]].grid(row=i, column=0, pady=1, padx=5)

        # binning
        self.inputs[self.binning] = LabelInput(
            parent=self.roi_frame,
            label=self.binning,
            input_class=ttk.Combobox,
            input_var=tk.StringVar(),
            input_args={"width": 5},
        )
        self.inputs[self.binning].grid(row=3, column=0, pady=5, padx=5)

        # Additional formatting
        #   Number of Pixels
        self.inputs["Width"].grid(pady=(10, 5))
        self.inputs["Width"].label.grid(padx=(0, 13))
        self.inputs["Height"].label.grid(padx=(0, 10))
        self.inputs["Binning"].label.grid(padx=(0, 6))
        #   FOV
        self.inputs["FOV_X"].grid(pady=(10, 5))
        self.inputs["FOV_X"].label.grid(padx=(0, 10))
        self.inputs["FOV_Y"].label.grid(padx=(0, 10))
        #   ROI Center
        self.inputs["Center_X"].grid(pady=(10, 5))
        self.inputs["Center_X"].label.grid(padx=(0, 10))
        self.inputs["Center_Y"].label.grid(padx=(0, 10))

    def get_variables(self):
        """Get Variables.

        This function returns a dictionary of all the variables that are tied to each
        widget name.

        The key is the widget name, value is the variable associated.

        Returns
        -------
        variables : dict
            Dictionary of all the variables tied to each widget.
        """
        variables = {}
        for key, widget in self.inputs.items():
            variables[key] = widget.get()
        return variables

    def get_widgets(self):
        """Get Widgets.

        This function returns the dictionary that holds the widgets.
        The key is the widget name, value is the LabelInput class that has all the data.

        Returns
        -------
        self.inputs : dict
            Dictionary of all the widgets in the frame.
        """
        return self.inputs

    def get_buttons(self):
        """Get Buttons.

        This function returns the dictionary that holds the buttons.
        The key is the widget name, value is the LabelInput class that has all the data.

        Returns
        -------
        self.buttons : dict
            Dictionary of all the buttons in the frame.
        """
        return self.buttons



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

#  Standard Library Imports
import logging

# Third Party Imports

# Local Imports
from navigate.model.devices.filter_wheel.filter_wheel_base import FilterWheelBase

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class SyntheticFilterWheel(FilterWheelBase):
    """SyntheticFilterWheel Class

    Attributes
    ----------
    comport : str
        Comport for communicating with the filter wheel, e.g., COM1.
    baudrate : int
        Baud rate for communicating with the filter wheel, e.g., 9600.
    filter_dictionary : dict
        Dictionary with installed filter names, e.g., filter_dictionary = {'GFP', 0}.
    number_of_filter_wheels : int
        Number of installed filter wheels.
    wheel_position : int
        Default filter wheel position
    wait_until_done_delay = float
        Duration of time to wait for a filter wheel change.
    wait_until_done = bool
        Flag for enabling the wait period for a filter wheel change.
    """

    def __init__(self, microscope_name, device_connection, configuration):
        super().__init__(microscope_name, device_connection, configuration)

    def filter_change_delay(self, filter_name):
        """Calculate duration of time necessary to change filter wheel positions

        Parameters
        ----------
        filter_name : str
            Name of the filter that we want to move to
        """
        pass

    def set_filter(self, filter_name, wait_until_done=True):
        """Change the filter wheel to the filter designated by the filter
        position argument.

        Parameters
        ----------
        filter_name : str
            Name of filter to move to.
        wait_until_done : bool
            Waits duration of time necessary for filter wheel to change positions.
        """
        pass

    def read(self, num_bytes):
        """Reads the specified number of bytes from the serial port.

        Parameters
        ----------
        num_bytes : int
            Number of bytes to read from the serial port.
        """
        pass

    def close(self):
        """Close the SyntheticFilterWheel.

        Sets the filter wheel to the Empty-Alignment position and then closes the port.
        """
        pass
